# coding: utf-8
"""
    SimpleGuinierAnimation.py

    Copyright (c) 2017-2018, Masatsuyo Takahashi, KEK-PF
"""
import numpy as np
import matplotlib.pyplot    as plt
import matplotlib.animation as animation
from mpl_toolkits.mplot3d   import Axes3D
from SimpleGuinier          import KNOT_NO_ADDITION
from SimpleGuinierScore     import NUM_SCORE_FACTORS, USE_FWD_CONSISTENCY

if USE_FWD_CONSISTENCY:
    SCORE_COLORS = [ 'blue', 'purple', 'magenta', 'pink' ]
else:
    SCORE_COLORS = [ 'blue', 'purple', 'magenta' ]
assert len(SCORE_COLORS) == NUM_SCORE_FACTORS

PLOT_BEST_HISTORY = False

class SimpleGuinierAnimation:
    def __init__( self, guinier, title, result_rg=None ):
        self.guinier    = guinier
        self.title      = title
        self.result_rg  = result_rg
        self.mpl_canvas_widget = None

    def destroy( self ):
        if self.mpl_canvas_widget is not None:
            print( '__del__' )
            self.mpl_canvas_widget.destroy()

    def draw( self, fig, interval=100, anim_iter_max=200, skip_first_stage=True ):
        if PLOT_BEST_HISTORY:
            ax1 = fig.add_subplot(231)
            ax2 = fig.add_subplot(232)
            ax3 = fig.add_subplot(233)
            ax5 = fig.add_subplot(235, projection='3d')
        else:
            ax1 = fig.add_subplot(131)
            ax2 = fig.add_subplot(132)
            ax3 = fig.add_subplot(133)

        x2      = self.guinier.x2
        log_y   = self.guinier.log_y

        if self.guinier.qrg_stop is None:
            for ax in [ax1, ax2, ax3]:
                ax.plot( x2, log_y, 'o', markersize=3 )
                xmin, xmax = ax.get_xlim()
                ymin, ymax = ax.get_ylim()
                ax.text( (0.7*xmin + 0.3*xmax), (ymin + ymax)/2, 'No Guinier interval' )
            return

        px2     = self.guinier.px2
        log_py  = self.guinier.log_py
        kx2     = self.guinier.kx2
        log_ky  = self.guinier.log_ky
        log_sy  = self.guinier.log_sy
        # log_y_max, log_y_min = np.percentile(log_y, [100,0])

        # TODO: in case self.guinier.guinier_start is None
        gn_min = log_y[self.guinier.qrg_stop]
        gn_max = log_y[self.guinier.guinier_start]
        if gn_min > gn_max:
            gn_min, gn_max = gn_max, gn_min

        ymin2 = gn_min - 1.0
        ymax2 = max( gn_max + 1.0, self.guinier.log_py + 0.2 )
        ax2.set_xlim( -0.0005, self.guinier.kx2[-3] * 1.1 )
        ax2.set_ylim( ymin2, ymax2 )
        xmax3 = self.guinier.x2[self.guinier.qrg_stop]
        ax3.set_xlim( max(-0.00025, -xmax3/6 ), xmax3 )
        ax3.set_ylim( min( self.guinier.guinier_y[1], (ymin2+ymax2)/2)-0.05, (self.guinier.guinier_y[0] + ymax2)/2 )

        knot_points = []
        for i, ax in enumerate([ax1, ax2, ax3]):
            if self.title is not None:
                ax.set_title( self.title )
            ax.set_xlabel( 'Q²' )
            ax.set_ylabel( 'Ln(I)' )
            ax.plot( x2, log_y, 'o', markersize=3 )
            ax.plot( x2, log_sy, color='green', alpha=0.5 )
            msize = 5 if i == 0 else 8

            point, = ax.plot( px2, log_py, 'o', color='red', markersize=msize )
            knot_points.append( point )
            point, = ax.plot( kx2, log_ky, 'o', color='yellow', markersize=msize )
            knot_points.append( point )
            if self.guinier.added_knot_type != KNOT_NO_ADDITION:
                point, = ax.plot( kx2[1], log_ky[1], 'o', color='pink', markersize=msize )
                knot_points.append( point )

        fig.tight_layout()

        if len(self.guinier.anim_ag_lines) > 0:
            ag_x, ag_y = self.guinier.anim_ag_lines[0]
        gstop_x = self.guinier.x2[self.guinier.anim_guinier_stop]
        anim_cand_len = len(self.guinier.anim_cand_list)
        if anim_cand_len > 0:
            start, stop, cx, cy, slope, intercept, score, score_vector, rg = self.guinier.anim_cand_list[0]
            endoints = [0, -1]
            cxe = cx[endoints]
            cye = slope * cxe + intercept
        else:
            score = 0
            score_vector = []

        lines = []
        glines = []
        clines = []
        rglines = []
        dlines = []
        bline_params = []
        ft_texts = []
        fline_bgs = []
        flines = []
        sline_bgs = []
        slines = []
        rg_texts = []
        final_rg_texts = []

        for ax in [ax2, ax3]:
            if len(self.guinier.anim_ag_lines) > 0:
                line, = ax.plot( ag_x, ag_y, color='orange', marker='o', markersize=8 )
                lines.append( line )
            gline, = ax.plot( [ gstop_x, gstop_x ], [ ymin2, ymax2 ], ':', color='gray', alpha=0.5 )
            glines.append( gline )

            if anim_cand_len > 0:
                cline, = ax.plot( cx, cy, color='yellow' )
                rgline, = ax.plot( cxe, cye, color='orange', marker='o', markersize=8 )
                clines.append( cline )
                rglines.append( rgline )

            dline, = ax.plot( self.guinier.guinier_x, self.guinier.guinier_y, color='cyan', marker='o', markersize=8 )
            dlines.append( dline )
            xmin, xmax = ax.get_xlim()
            ymin, ymax = ax.get_ylim()
            bx = 0.5*xmin + 0.5*xmax
            by = 0.1*ymin + 0.9*ymax
            bsize = (xmax - xmin)*0.4
            byoffset = (ymax-ymin)*0.05
            bline_params.append( [bx, by, bsize, byoffset] )

            # start-stop
            ax.text( bx-bsize*0.34, by, 'From-To:', verticalalignment='center' )
            ft_text = ax.text( bx, by, '', verticalalignment='center' )
            ft_texts.append( ft_text )

            # Factors
            by_f = by - byoffset
            ax.text( bx-bsize*0.3, by_f, 'Factors:', verticalalignment='center' )
            by_f_ = [ by_f, by_f ]
            fline_bg, = ax.plot( [bx, bx+bsize], by_f_, color='white', linewidth=10, solid_capstyle='butt' )
            fline_bgs.append( fline_bg )
            bxstt = bx
            for s, score in enumerate(score_vector):
                bxend = bxstt + bsize*score
                bx_ = [bxstt, bxend]
                fline, = ax.plot( bx_, by_f_, color=SCORE_COLORS[s], linewidth=10, solid_capstyle='butt' )
                bxstt = bxend
                flines.append( fline )

            # Score
            by_s = by - byoffset*2
            ax.text( bx-bsize*0.25, by_s, 'Score:', verticalalignment='center' )
            by_s_ = [by_s, by_s]
            sline_bg, = ax.plot( [bx, bx+bsize], by_s_, color='white', linewidth=10, solid_capstyle='butt' )
            sline_bgs.append( sline_bg )
            sline, = ax.plot( [bx, bx], by_s_, color='green', linewidth=10, solid_capstyle='butt' )
            slines.append( sline )

            # Rg
            by_rg = by - byoffset*3
            ax.text( bx-bsize*0.13, by_rg, 'Rg:', verticalalignment='center' )
            rg_text = ax.text( bx, by_rg, '', verticalalignment='center' )
            rg_texts.append( rg_text )

            # final Rg
            by_rg = by - byoffset*4
            ax.text( bx-bsize*0.34, by_rg, 'Final Rg:', verticalalignment='center' )
            rg_text = ax.text( bx, by_rg, '', verticalalignment='center' )
            final_rg_texts.append( rg_text )

        if PLOT_BEST_HISTORY:
            best_x = []
            best_y = []
            best_z = []
            for k, rec in self.guinier.best_history_dict.items():
                v, rg = rec
                size_score = v[0]
                print( '(%3d, %3d) rg=%.3g, size_score=%.4g' % ( k[0], k[1], rg, size_score ) )
                best_x.append( k[0] )
                best_y.append( k[1] )
                best_z.append( size_score )

            ax5.plot( best_x, best_y, best_z )

        def init():
            for line in lines:
                line.set_ydata(np.ma.array( ag_x, mask=True))
            for gline in glines:
                gline.set_ydata(np.ma.array( [ gstop_x, gstop_x ], mask=True))
            for cline in clines:
                cline.set_xdata(np.ma.array( cx, mask=True))
                cline.set_ydata(np.ma.array( cy, mask=True))
            for rgline in rglines:
                rgline.set_xdata(np.ma.array( cxe, mask=True))
                rgline.set_ydata(np.ma.array( cye, mask=True))
            for dline in dlines:
                dline.set_xdata(np.ma.array( self.guinier.guinier_x, mask=True))
                dline.set_ydata(np.ma.array( self.guinier.guinier_y, mask=True))
            for sline in slines:
                sline.set_xdata(np.ma.array( [bx, bx], mask=True))
                sline.set_ydata(np.ma.array( by_s_, mask=True))
            return tuple( lines + glines + clines + rglines + slines + dlines )

        num_iter_cand = min( anim_iter_max, len( self.guinier.anim_cand_list ) )
        result_rg = self.guinier.Rg if self.result_rg is None else self.result_rg

        def draw_factors( a_score_vector, a_score ):
            for k in range(2):
                bx, by, bsize, byoffset = bline_params[k]
                by_f = by - byoffset
                by_f_ = [ by_f, by_f ]

                # score factors
                bxstt = bx
                for s, score in enumerate(a_score_vector):
                    bxend = bxstt + bsize*score
                    bx_ = [bxstt, bxend]
                    fline = flines[NUM_SCORE_FACTORS*k + s]
                    fline.set_xdata( bx_ )
                    fline.set_ydata( by_f_ )
                    bxstt = bxend

                # score
                bx_s_ = [ bx, bx + bsize*a_score ]
                by_s = by - byoffset*2
                by_s_ = [ by_s, by_s ]

                sline = slines[k]
                sline.set_xdata( bx_s_ )
                sline.set_ydata( by_s_ )

        def animate( i ):
            if i < len(self.guinier.anim_ag_lines):
                ag_x, ag_y = self.guinier.anim_ag_lines[i]
                for line in lines:
                    line.set_xdata( ag_x )
                    line.set_ydata( ag_y )
                return tuple( lines + knot_points )
            else:
                j = i - len(self.guinier.anim_ag_lines)
                if j == 0:
                    for gline in glines:
                        gline.set_ydata( [ ymin2, ymax2 ] )
                    return tuple( glines )
                elif j < num_iter_cand:

                    # draw also the final guinier line in early stages
                    for dline in dlines:
                        dline.set_xdata( self.guinier.guinier_x )
                        dline.set_ydata( self.guinier.guinier_y )

                    start, stop, cx, cy, slope, intercept, score, score_vector, rg = self.guinier.anim_cand_list[j]
                    for cline in clines:
                        cline.set_xdata( cx )
                        cline.set_ydata( cy )
                    cxe = cx[endoints]
                    cye = slope * cxe + intercept
                    for rgline in rglines:
                        rgline.set_xdata( cxe )
                        rgline.set_ydata( cye )

                    for ft_text in ft_texts:
                        ft_text.set_text( '%3d-%3d' % ( start, stop ) )

                    draw_factors(score_vector, score)

                    for rg_text in rg_texts:
                        rg_text.set_text( '%.3g' % rg )

                    for rg_text in final_rg_texts:
                        rg_text.set_text( '%.3g' % result_rg )

                    return tuple( dlines + glines + clines + rglines + ft_texts + fline_bgs + flines + slines + rg_texts + final_rg_texts + knot_points )
                else:
                    for dline in dlines:
                        dline.set_xdata( self.guinier.guinier_x )
                        dline.set_ydata( self.guinier.guinier_y )

                    if anim_cand_len > 0:
                        for ft_text in ft_texts:
                            ft_text.set_text( '%3d-%3d' % ( self.guinier.guinier_start, self.guinier.guinier_stop ) )
                        draw_factors(self.guinier.score_vector, self.guinier.score)

                    for rg_text in rg_texts:
                        rg_text.set_text( '%.3g' % result_rg )

                    for rg_text in final_rg_texts:
                        rg_text.set_text( '%.3g' % result_rg )

                    return tuple( dlines + ft_texts + fline_bgs + flines + slines + rg_texts + final_rg_texts + knot_points )

        num_pause = 50
        print( 'num_iter_cand=', num_iter_cand )
        num_iter_first = len(self.guinier.anim_ag_lines)
        if skip_first_stage:
            range_ = np.arange( num_iter_first, num_iter_first+num_iter_cand+num_pause )
        else:
            range_ = np.arange(1, num_iter_first+num_iter_cand+num_pause )
        self.anim = animation.FuncAnimation(fig, animate, range_, init_func=init,
                    interval=interval, blit=True)



# coding: utf-8
"""
    SimpleGuinierMain.py

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF
"""
import numpy as np
import re
from mpl_toolkits.mplot3d       import proj3d
from molass_legacy.KekLib.OurTkinter                 import Tk
from molass_legacy.KekLib.TkUtils                    import adjusted_geometry, split_geometry
from molass_legacy.KekLib.TkCustomWidgets            import FileEntry, FolderEntry
from SimpleGuinierAnalyzer      import SimpleGuinierAnalyzer
from CanvasFrame                import CanvasFrame
from molass_legacy.KekLib.OurMatplotlib              import NavigationToolbar
from CanvasDialog               import CanvasDialog
from DataUtils                  import get_pytools_folder
from SerialData                 import SerialData, find_conc_files
from molass_legacy._MOLASS.SerialSettings             import get_setting
from SimpleGuinierGuiUtils      import window_title, CheckbuttonFrame, TypeInfoPanel, FolderInfoPanel
from SimpleGuinierFolderAnalyzer    import SimpleGuinierFolderAnalyzer

class SimpleGuinierMain( Tk.Toplevel ):
    def __init__( self, parent, tester_log=None ):
        self.parent = parent
        Tk.Toplevel.__init__( self, parent )
        self.withdraw()
        self.title( window_title )

        frame0 = Tk.Frame( self )
        frame0.pack()
        frame01 = Tk.Frame( frame0 )
        frame01.pack( side=Tk.LEFT, padx=10 )
        frame02 = Tk.Frame( frame0 )
        frame02.pack( side=Tk.LEFT, padx=10 )
        frame03 = Tk.Frame( frame0 )
        frame03.pack( side=Tk.LEFT, padx=10 )

        label = Tk.Label( frame01, text='Tester Log: ' )
        label.pack( side=Tk.LEFT )

        self.tester_log = Tk.StringVar()
        entry   = FileEntry( frame01, textvariable=self.tester_log, width=50, on_entry_cb=self.on_entry_tester_log )
        entry.pack( side=Tk.LEFT )

        label = Tk.Label( frame02, text='Data Folder: ' )
        label.pack( side=Tk.LEFT )

        self.pytools_folder = get_pytools_folder()
        self.data_folder = Tk.StringVar()
        self.data_folder.set( self.pytools_folder + '/Data' )
        entry   = FolderEntry( frame02, textvariable=self.data_folder, width=50 )
        entry.pack( side=Tk.LEFT )

        label = Tk.Label( frame03, text='Report Folder: ' )
        label.pack( side=Tk.LEFT )

        self.report_folder = Tk.StringVar()
        self.report_folder.set( self.pytools_folder + '/TODO/guinier-results/reports-1_1_5' )
        entry   = FolderEntry( frame03, textvariable=self.report_folder, width=50 )
        entry.pack( side=Tk.LEFT )

        frame1 = Tk.Frame( self )
        frame1.pack()
        self.canvas = CanvasFrame( frame1, figsize=(12, 8) )
        self.canvas.pack( side=Tk.LEFT )
        self.toolbar = NavigationToolbar( self.canvas.mpl_canvas, self.canvas )
        self.create_popup_menu()

        frame12 = Tk.Frame( frame1 )
        frame12.pack( side=Tk.RIGHT, fill=Tk.BOTH  )
        label = Tk.Label( frame12, text='Control Panel' )
        label.pack( anchor=Tk.N )

        view_init_btn = Tk.Button( frame12, text='view init', command=self.view_init )
        view_init_btn.pack()

        self.cb_frame = CheckbuttonFrame( frame12 )
        self.cb_frame.pack()

        redraw_btn   = Tk.Button( frame12, text='redraw', command=self.redraw )
        redraw_btn.pack()

        type_info_btn = Tk.Button( frame12, text='type help', command=self.show_type_info )
        type_info_btn.pack()

        folder_info_btn = Tk.Button( frame12, text='folder help', command=self.show_folder_info )
        folder_info_btn.pack()

        hist_btn = Tk.Button( frame12, text='histogram', command=self.show_hist )
        hist_btn.pack()

        folder_frame = Tk.Frame( frame12 )
        folder_frame.pack()
        folder_frame1 = Tk.Frame( folder_frame )
        folder_frame1.pack()
        label = Tk.Label( folder_frame1, text='Folder No:' )
        label.pack( side=Tk.LEFT )
        self.folder_no = Tk.IntVar()
        entry = Tk.Entry( folder_frame1, textvariable=self.folder_no, width=3, justify=Tk.CENTER )
        entry.pack( side=Tk.LEFT )

        show_folder_btn = Tk.Button( folder_frame, text='draw folder', command=self.draw_folder )
        show_folder_btn.pack()

        self.update()
        self.deiconify()
        self.geometry( adjusted_geometry(self.geometry()) )
        self.update()
        self.protocol( "WM_DELETE_WINDOW", self.quit )
        if tester_log is not None:
            self.tester_log.set( tester_log )
            self.on_entry_tester_log()

    def quit( self ):
        self.parent.quit()
        self.destroy()

    def on_entry_tester_log( self ):
        tester_log = self.tester_log.get()
        print( 'tester_log=', tester_log )
        self.analyzer = SimpleGuinierAnalyzer( tester_log )
        self.num_folders = len( self.analyzer.folder_infos )
        self.canvas.draw( self.analyzer.plot_results )
        self.canvas.fig.canvas.mpl_connect( 'button_press_event', self.on_button_press )
        self.canvas.fig.canvas.mpl_connect( 'draw_event', self.on_draw )
        self.axes = [ self.analyzer.ax1 ]
        self.update_xyr( force=True )
        self.compute_projected_xy()

    def update_xyr( self, force=False ):
        if force or self.cb_frame.changed:
            index   = self.cb_frame.make_index( self.analyzer.z )
            self.x  = self.analyzer.x[index]
            self.y  = self.analyzer.y[index]
            self.r  = self.analyzer.r[index]
            print( 'len(analyzer.x)', len(self.analyzer.x),'len(self.x)=', len(self.x) )

    def compute_projected_xy( self ):
        self.update_xyr()

        ax = self.axes[0]
        # print( ax.get_proj() )
        projection = ax.get_proj()
        x, y, _ = proj3d.proj_transform( self.x, self.y, self.r, projection )
        xy = np.vstack( [x, y] ).T
        ret = ax.transData.transform( xy )
        print( 'ret.shape=', ret.shape )
        self.x_scr = ret[:,0]
        self.y_scr = ret[:,1]
        if False:
            for i in range(5):
                print( 'point[%d]=' % i, ( self.x[i], self.y[i], self.r[i] ) )
                # print( 'transformed point[%d]=' % i, (x[i], y[i]) )
                print( 'transformed screen point[%d]=' % i, (self.x_scr[i], self.y_scr[i]) )

    def create_popup_menu( self ):
        self.popup_menu = Tk.Menu( self, tearoff=0 )
        self.popup_menu.add_command( label='Animation', command=self.draw_folder )

    def on_button_press( self, event ):
        if event.button == 3:
            self.on_right_button_press( event )
            return

    def on_right_button_press( self, event ):
        print( 'on_right_button_press' )

        if event.inaxes not in self.axes:
            return

        print( 'x=', event.x )
        print( 'y=', event.y )
        w, h, x, y = split_geometry( self.canvas.mpl_canvas_widget.winfo_geometry() )
        h_ = h
        # print( 'self.mpl_canvas_widget.winfo_geometry()=', (w, h, x, y) )
        w, h, x, y = split_geometry( self.winfo_geometry() )
        # print( 'self.winfo_geometry()=', (w, h, x, y) )
        # TODO: better way to get the mouse cursor position

        folder_no = self.guess_pointed_folder_no( event )
        if folder_no is None:
            return

        self.folder_no.set( folder_no )
        self.popup_menu.entryconfig( 0, label="draw folder %d" % folder_no )
        self.popup_menu.post( int( x + event.x + 10 ), int( y + h_ - event.y + 50 ) )

    def on_draw( self, event ):
        print( 'on_draw' )
        self.compute_projected_xy()

    def guess_pointed_folder_no( self, event ):
        dist = ( self.x_scr - event.x )**2 + ( self.y_scr - event.y )**2
        i = np.argmin( dist )
        folder_no = int( self.y[i] )
        print( 'y[%d]=%g' % ( i, folder_no ) )
        return folder_no

    def view_init( self ):
        self.analyzer.ax1.view_init()
        self.canvas.mpl_canvas.show()

    def redraw( self ):
        draw_flags = [ var.get() for var in self.cb_frame.cb_vars ]

        def redraw_closure( fig ):
            self.analyzer.plot_redraw( fig, draw_flags=draw_flags )

        self.canvas.draw( redraw_closure )
        # self.canvas.mpl_canvas.show()

    def draw_folder( self, folder_no=None ):
        if folder_no is None:
            folder_no = self.folder_no.get()
        print( 'draw_folder: folder_no=', folder_no )
        analyzer = self.analyzer
        folder_path = self.rename_folder_path( analyzer.folder_paths[folder_no] )
        folder_info = analyzer.folder_infos[folder_no]
        index = analyzer.y == folder_no
        x = analyzer.x[index]
        z = analyzer.z[index]
        r = analyzer.r[index]
        fa = SimpleGuinierFolderAnalyzer( self, folder_no, folder_path, folder_info, x, z, r, analyzer.type_names )
        fa.show()

    def rename_folder_path( self, path ):
        nodes = path.split( '/' )
        i = nodes.index( 'Data' )
        new_path = self.pytools_folder + '/' + '/'.join( nodes[i:] )
        return new_path

    def show_type_info( self ):
        dialog = TypeInfoPanel( self )
        dialog.show( self.analyzer.type_names )

    def show_folder_info( self ):
        dialog = FolderInfoPanel( self )
        dialog.show( self.analyzer.folder_paths )

    def show_hist( self ):
        dialog = CanvasDialog( "Debug", parent=self )
        dialog.show( self.draw_hist, figsize=(9,8), toolbar=True )

    def draw_hist( self, fig ):
        ax = fig.add_subplot(111)
        ax.set_title( 'Histogram of Peak Process Types' )
        bins = np.arange(18) - 0.5
        # print( 'bins=', bins )
        # bins = 17
        ax.hist( self.analyzer.z, bins=bins )
        ax.xaxis.set_ticks(np.arange(17))
        fig.tight_layout()



# coding: utf-8
"""
    WLS_Basics.py

    Copyright (c) 2016-2018, Masatsuyo Takahashi, KEK-PF
"""
import numpy            as np

class Result:
    def __init__( self, **entries ): 
        self.__dict__.update(entries)

    def cov_params( self ):
        return self.cov_params_

class OLS:
    def __init__( self, y, x ):
        assert( len( y.shape ) == 1 )
        assert( len( x.shape ) == 1 )
        assert( y.shape[0] == x.shape[0] )

        self.N = N = y.shape[0]
        shape_ = ( N, 1 )
        self.A          = np.hstack( [ np.ones( shape_ ), np.reshape( x, shape_ ) ] )   # add constant
        # print( self.A )
        self.b          = np.reshape( y, shape_ )

    def fit( self ):
        At      = np.transpose( self.A )
        AtA_inv = np.linalg.inv( np.dot( At, self.A ) )
        params  = np.dot( np.dot( AtA_inv, At ), self.b )
        # print( params )
        e       = self.b - np.dot( self.A, params )
        S       = np.sum( e**2 )
        sigma2  = S / ( self.N - 2 )
        cov_params = AtA_inv * sigma2
        # print( cov_params )
        return Result(
                params=np.reshape( params, ( params.shape[0], ) ),
                cov_params_=cov_params,
                )

class WLS:
    def __init__( self, y, x, w ):
        assert( len( y.shape ) == 1 )
        assert( len( x.shape ) == 1 )
        assert( len( w.shape ) == 1 )
        assert( y.shape[0] == x.shape[0] )
        assert( y.shape[0] == w.shape[0] )

        self.x  = x
        self.y  = y
        self.N  = N = y.shape[0]
        shape_ = ( N, 1 )
        self.A          = np.hstack( [ np.ones( shape_ ), np.reshape( x, shape_ ) ] )   # add constant
        # print( self.A )
        self.b          = np.reshape( y, shape_ )
        # w = w / np.sum( w )
        self.w          = w
        self.W          = np.diag( w )

    def fit( self ):
        At      = np.transpose( self.A )
        AtW     = np.dot( At, self.W )
        AtWA_inv = np.linalg.inv( np.dot( AtW, self.A ) )
        AtWb    = np.dot( AtW, self.b )
        params  = np.dot( AtWA_inv, AtWb )
        # print( params )
        """
            As for the covariance matrix of params, see
            Linear least squares (mathematics)
            https://en.wikipedia.org/wiki/Linear_least_squares_%28mathematics%29#Weighted_linear_least_squares
        """
        e       = ( self.b - np.dot( self.A, params ) ).squeeze()
        ee = e**2
        print( 'e_=', e[0:5] )
        print( 'e =', 1/np.sqrt( self.w[0:5] ) )
        S       = np.inner( self.w, ee )
        sigma2  = S / ( self.N - 2 )
        print( 'S=', S, ', N=', self.N, ', sigma2=', sigma2 )
        cov_params = AtWA_inv * sigma2

        ssr = np.sum(ee)
        # TODO: this ssr value is differenct from that of statsmodels.api

        return Result(
                params=np.reshape( params, ( params.shape[0], ) ),
                cov_params_=cov_params,
                ssr=ssr,
                )



"""
    CharFunc/PdfCfStudy.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""

import numpy as np
import matplotlib.pyplot as plt

def study_egh():
    from molass_legacy.Models.EGH import egh
    fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18,5))
    ax1.set_title("PDF Domain")
    ax2.set_title("CF Real Domain")
    ax3.set_title("CF Imaginary Domain")
    x = np.arange(300)
    y1 = egh(x, 1, 120, 20, 0)
    y2 = egh(x, 1, 150, 30, 10)
    ty = y1 + y2
    tz = np.fft.fft(ty)
    w = np.fft.fftfreq(len(ty))
    u = np.fft.ifft(tz)
    z1 = np.fft.fft(y1)
    z2 = np.fft.fft(y2)
    ax1.plot(x, ty, label="data")
    ax1.plot(x, y1, ":", label="component-1")
    ax1.plot(x, y2, ":", label="component-2")

    # ax2.plot(w, np.real(tz), "o", markersize=1)
    # ax3.plot(w, np.imag(tz), "o", markersize=1)
    ax2.plot(w, np.real(z1), "o", markersize=1)
    ax2.plot(w, np.real(z2), "o", markersize=1)
    ax3.plot(w, np.imag(z1), "o", markersize=1)
    ax3.plot(w, np.imag(z2), "o", markersize=1)

    ax1.plot(x, u, ":", color="cyan", label="inverse FFT")
    v = np.fft.ifft(z1 + z2)
    ax1.plot(x, v, ":", color="red", label="inverse FFT of sum")

    ax1.legend()
    fig.tight_layout()
    plt.show()

def study_phi():
    from SecTheory.BasicModels import single_pore_pdf
    from SecTheory.SecCF import simple_phi, shifted_phi

    fig, (ax1, ax2, ax3, ax4) = plt.subplots(ncols=4, figsize=(20,4))
    ax1.set_title("PDF Domain")
    ax2.set_title("CF Real Domain")
    ax3.set_title("CF Imaginary Domain")

    x = np.arange(21, 321)
    x0 = 20
    npi = 100
    tpi = 1
    y1 = single_pore_pdf(x, npi, tpi)

    x = np.arange(21, 321)
    y1 = single_pore_pdf(x, npi, tpi)
    y2 = single_pore_pdf(x - x0, npi, tpi)
    ty = y1 + y2
    tz = np.fft.fft(ty)
    w = np.fft.fftfreq(len(ty))
    u = np.fft.ifft(tz)
    z1 = np.fft.fft(y1)
    z2 = np.fft.fft(y2)

    z1_ = simple_phi(w, npi, tpi)
    z2_ = shifted_phi(w, npi, tpi, -x0)
    u1_ = np.fft.ifft(z1_)
    u2_ = np.fft.ifft(z2_)

    ax1.plot(x, ty, label="data")
    ax1.plot(x, y1, ":", label="component-1")
    ax1.plot(x, y2, ":", label="component-2")

    ax1.plot(x, u1_, color="red", alpha=0.5, label="component-1 inverse FFT")
    ax1.plot(x, u2_, color="green", alpha=0.5, label="component-2 inverse FFT")

    # ax2.plot(w, np.real(tz), "o", markersize=1)
    # ax3.plot(w, np.imag(tz), "o", markersize=1)
    ax2.plot(w, np.real(z1), "o", markersize=1)
    ax2.plot(w, np.real(z2), "o", markersize=1)

    ax3.plot(w, np.imag(z1), "o", markersize=1)
    ax3.plot(w, np.imag(z2), "o", markersize=1)

    ax1.plot(x, u, ":", color="cyan", label="inverse FFT")
    v = np.fft.ifft(z1 + z2)
    ax1.plot(x, v, ":", color="red", label="inverse FFT of sum")

    ax1.legend()

    ax4.plot(w, "o", markersize=1)
    ax4.plot(0, w[0], "o", color="red")
    ax4.plot(len(w)-1, w[-1], "o", color="blue")

    fig.tight_layout()
    plt.show()

def study_phi_with_Witkovsky():
    from SecTheory.BasicModels import single_pore_pdf
    from SecTheory.SecCF import simple_phi, shifted_phi
    from SecTheory.SecPDF import BidirectFft

    fig, (ax1, ax2, ax3, ax4) = plt.subplots(ncols=4, figsize=(20,4))
    ax1.set_title("PDF Domain")
    ax2.set_title("CF Real Domain")
    ax3.set_title("CF Imaginary Domain")

    x = np.arange(21, 321)
    x0 = 20
    npi = 100
    tpi = 1
    y1 = single_pore_pdf(x, npi, tpi)

    x = np.arange(21, 321)
    y1 = single_pore_pdf(x, npi, tpi)
    y2 = single_pore_pdf(x - x0, npi, tpi)
    ty = y1 + y2

    bifft = BidirectFft(x)

    tz = bifft.compute_z(ty)
    w = bifft.get_w()
    u = bifft.compute_y(tz)

    z1 = bifft.compute_z(y1)
    z1_ = simple_phi(w, npi, tpi)
    u1_ = bifft.compute_y(z1_)

    z2 = bifft.compute_z(y2)
    z2_ = shifted_phi(w, npi, tpi, x0)
    u2_ = bifft.compute_y(z2_)

    ax1.plot(x, ty, label="data")
    ax1.plot(x, y1, ":", label="component-1")
    ax1.plot(x, y2, ":", label="component-2")

    ax1.plot(x, u1_, color="red", alpha=0.5, label="component-1 inverse FFT")
    ax1.plot(x, u2_, color="green", alpha=0.5, label="component-2 inverse FFT")

    # ax2.plot(w, np.real(tz), "o", markersize=1)
    # ax3.plot(w, np.imag(tz), "o", markersize=1)
    ax2.plot(w, np.real(z1), "o", markersize=1)
    ax2.plot(w, np.real(z2), "o", markersize=1)

    ax3.plot(w, np.imag(z1), "o", markersize=1)
    ax3.plot(w, np.imag(z2), "o", markersize=1)

    ax1.plot(x, u, ":", color="cyan", label="inverse FFT")
    v = bifft.compute_y(z1 + z2)
    ax1.plot(x, v, ":", color="red", label="inverse FFT of sum")

    ax1.legend()

    ax4.plot(w, "o", markersize=1)
    ax4.plot(0, w[0], "o", color="red")
    ax4.plot(len(w)-1, w[-1], "o", color="blue")

    fig.tight_layout()
    plt.show()


def study_Gil_Pelaez():
    mu = 100
    sigma = 10
    def f(x):
        return 1/(sigma*np.sqrt(2*np.pi)) * np.exp(-(x - mu)**2 / (2 * sigma**2))

    x = np.arange(300)


if __name__ == "__main__":
    import sys
    import os
    sys.path.append(os.path.dirname(os.path.abspath(__file__)) + "/../")

    # study_egh()
    #study_phi()
    study_phi_with_Witkovsky()
    # study_Gil_Pelaez()


# coding: utf-8
"""
    PatchUtils.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""

def make_area_points(x, y, bottomline=None):
    if bottomline is None:
        return [(x[0],0)] + list(zip(x,y)) + [(x[-1], 0)]
    else:
        r = list(zip(x,bottomline))
        r.reverse()
        return list(zip(x,y)) + r



"""

    GuiSettings.py

    Copyright (c) 2016-2025, SAXS Team, KEK-PF

"""
from molass_legacy.KekLib.OurTkinter import Tk
from molass_legacy._MOLASS.SerialSettings import get_setting

class GuiSettingsMenu(Tk.Menu):
    def __init__(self, parent, menubar ):
        self.parent = parent

        Tk.Menu.__init__(self, menubar, tearoff=0 )
        menubar.add_cascade( label="Settings", menu=self )
        self.add_command( label="Basic Settings", command=self.show_settings_dialog )
        self.add_command( label="Specialist Options", command=self.show_specialist_dialog )
        self.add_command( label="Check Environment", command=self.check_env)

    def show_settings_dialog(self, debug=True):
        if debug:
            from importlib import reload
            import molass_legacy.SerialAnalyzer.SettingsDialog
            reload(molass_legacy.SerialAnalyzer.SettingsDialog)
        from molass_legacy.SerialAnalyzer.SettingsDialog import SettingsDialog
        while True:
            dialog = SettingsDialog( self.parent, 'Settings' )
            if dialog.applied is None:
                # i.e., some errors exist
                # see SettingsDialog.py
                continue
            else:
                # i.e., dialog.applied is True (without any error) or False
                break

        if dialog.applied:
            self.parent.after(500, self.parent.re_prepare_serial_data)

    def show_specialist_dialog(self, debug=True):
        if debug:
            from importlib import reload
            import molass_legacy.SerialAnalyzer.SpecialistOptions
            reload(molass_legacy.SerialAnalyzer.SpecialistOptions)
        from molass_legacy.SerialAnalyzer.SpecialistOptions import SpecialistOptionsDialog
        prev_use_xray_conc = get_setting( 'use_xray_conc' )
        prev_apply_backsub = get_setting( 'apply_backsub' )
        dialog = SpecialistOptionsDialog( self.parent, 'Specialist Options' )
        dialog.show()
        if dialog.applied:
            self.parent.update_uv_folder_entry_state()
            self.parent.update_analysis_button_state()
            if ( get_setting( 'use_xray_conc' ) != prev_use_xray_conc
                or get_setting( 'apply_backsub' ) != prev_apply_backsub
                ):
                self.parent.loader.reset_current_status()

    def check_env(self):
        print('check_env')
        from molass_legacy.Env.CheckEnvDialog import CheckEnvDialog
        dialog = CheckEnvDialog(self.parent)
        dialog.show()



"""
    Simulative.ApproxBoundary.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
from SecTheory.BasicModels import robust_single_pore_pdf
from molass_legacy.Models.Stochastic.LognormalPoreFunc import lognormal_pore_pdf
from molass_legacy.Models.Stochastic.LognormalUtils import compute_mode, compute_stdev, compute_mu_sigma

def compute_lnp_curves(x, N, T, me, mp, mu, sigma, t0, rgs, scales):
    cy_list = []
    for rg, scale in zip(rgs, scales):
        cy = lognormal_pore_pdf(x, N, T, me, mp, mu, sigma, rg, t0)*scale
        cy_list.append(cy)
    ty = np.sum(cy_list, axis=0)
    return cy_list, ty

def compute_mnp_curves(x, N, T, me, mp, poresize, t0, rgs, scales):
    cy_list = []
    for rg, scale in zip(rgs, scales):
        rho = min(1, rg/poresize)
        ni_ = N * (1 - rho)**me
        ti_ = T * (1 - rho)**mp
        cy = robust_single_pore_pdf(x - t0, ni_, ti_)*scale
        cy_list.append(cy)
    ty = np.sum(cy_list, axis=0)
    return cy_list, ty

def compute_bounadry_ratio(x, N, T, t0, me, mp, poresize, rgs, scales, stdev_ratios=None):
    if stdev_ratios is None:
        stdev_ratios = np.arange(0.1, 1.001, 0.01)
    ty_mnp = compute_mnp_curves(x, N, T, me, mp, poresize, t0, rgs, scales)[1]
    area = np.sum(ty_mnp)
    ok_ratio = np.nan

    for j, ratio in enumerate(stdev_ratios):
        stdev = ratio * poresize
        mu_, sigma_ = compute_mu_sigma(poresize, stdev)
        ty_lnp = compute_lnp_curves(x, N, T, me, mp, mu_, sigma_, t0, rgs, scales)[1]
        dev_ratio = np.sum(np.abs(ty_lnp - ty_mnp))/area
        if dev_ratio > 0.3:
            break
        ok_ratio = ratio

    return ok_ratio

def compute_approx_boundary(x, y, params, rgs, quickly=True):
    N, T, t0, me, mp, mu, sigma = params[0:7]
    scales = params[7:]

    max_rg = np.max(rgs)
    rv = np.arange(int(round(max_rg + 1,-1)), 500, 10 if quickly else 1)
    stdev_ratio_list = np.arange(0.1, 1.001, 0.01 if quickly else 0.001)
    results = []
    for i, poresize in enumerate(rv):
        print(poresize)
        ok_ratio = compute_bounadry_ratio(x, N, T, t0, me, mp, poresize, rgs, scales, stdev_ratio_list)
        results.append(ok_ratio)

    if False:
        print("mu, sigma=", mu, sigma)
        np.savetxt("temp/boundary.dat", np.array([rv, results]).T)    

    results = np.array(results)
    return rv, results

def decide_allowance_color(rv, bv, mode, stdev_ratio):
    from bisect import bisect_right
    from scipy.interpolate import UnivariateSpline

    valid = np.isfinite(bv)    # NaN is harmful for UnivariateSpline
    i = bisect_right(rv, mode)
    if valid[i] and i+1 < len(valid) and valid[i+1]:
        spline = UnivariateSpline(rv[valid], bv[valid], s=0)
        boundary = spline(mode)
        color = "red" if stdev_ratio > boundary else "blue"
    else:
        color = "gray"
    return color

def plot_boundary_impl(ax2, rv, bv, mu, sigma):        
    ax2.set_title("Stdev Ratio vs. Pore Size", fontsize=16)
    ax2.set_xlabel(r"Pore Size ($\AA$)", fontsize=14)
    ax2.set_ylabel("Stdev Ratio", fontsize=14)
    ax2.plot(rv, bv, "-", linewidth=3, label="Approximation Bounadry")
    ymin, ymax = ax2.get_ylim()
    zeros = np.zeros(len(rv))
    tops = np.ones(len(rv))*ymax
    ax2.fill_between(rv, zeros, bv, color='green', alpha=0.2)
    ax2.fill_between(rv, bv, tops, color='pink', alpha=0.2)
    mode = compute_mode(mu, sigma)
    ax2.axvline(mode, color='blue', linestyle=':', label="Estimated Pore Size")
    stdev_ratio = compute_stdev(mu, sigma)/mode
    color = decide_allowance_color(rv, bv, mode, stdev_ratio)
    ax2.plot(mode, stdev_ratio, "o", color=color, markersize=10, label="Estimated Stdev Ratio")
    ax2.legend()
    return color

def plot_boundary(x, y, params, rgs, rv, bv, parent=None, fig_file=None):
    N, T, t0, me, mp, mu, sigma = params[0:7]
    scales = params[7:]

    cy_list, ty = compute_lnp_curves(x, N, T, me, mp, mu, sigma, t0, rgs, scales)
    if y is None:
        y = ty

    with plt.Dp(parent=parent, window_title="Demo", button_spec=["OK", "Cancel"]):
        fig, (ax1,ax2) = plt.subplots(ncols=2, figsize=(12,5))
        fig.suptitle("Monopore Approximation Boundary for %s" % get_in_folder(), fontsize=20)
        ax1.set_title("Experiment Data", fontsize=16)
        ax1.set_xlabel("Time (Frames)", fontsize=14)
        ax1.set_ylabel("Intensity", fontsize=14)
        ax1.plot(x, y)
        if cy_list is not None:
            for cy in cy_list:
                ax1.plot(x, cy, ":")
        color = plot_boundary_impl(ax2, rv, bv, mu, sigma)
        fig.tight_layout()
        if fig_file is None:
            plt.show()
        else:
            plt.show(block=False)
            fig.savefig(fig_file)
    return color

def demo_impl(x, y, params, rgs, parent=None, quickly=True, fig_file=None):
    rv, bv = compute_approx_boundary(x, y, params, rgs, quickly=quickly)
    color = plot_boundary(x, y, params, rgs, rv, bv, parent=parent, fig_file=fig_file)
    return color

def compute_boundary_impl(lrf_src, quickly=True, parent=None, fig_file=None):
    guess_info = lrf_src.guess_lnpore_params(return_rgs=True)
    x = lrf_src.xr_x
    y = lrf_src.xr_y
    params, rgs = guess_info
    rv, bv = compute_approx_boundary(x, y, params, rgs, quickly=quickly)
    color = plot_boundary(x, y, params, rgs, rv, bv, parent=parent, fig_file=fig_file)
    ret_params = np.concatenate(guess_info)
    return color, ret_params


# coding: utf-8
"""
    CustomNavigationToolbar.py

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF
"""
from molass_legacy.KekLib.OurMatplotlib  import NavigationToolbar

class CustomNavigationToolbar( NavigationToolbar ):
    def __init__( self, canvas, window ):
        NavigationToolbar.__init__( self, canvas, window )

    """
        NavigationToolbar.mouse_move を override
    """
    def mouse_move(self, event):
        self._set_cursor(event)
        self.set_message(self.mode)



"""
    Optimizer.ResultAnimation.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import os
from time import sleep
from matplotlib.animation import ArtistAnimation
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting

def show_result_animation_impl(selector_dialog, debug=True):
    optimizer_folder = get_setting('optimizer_folder')
    joblist_folder = selector_dialog.folder.get()
    print("show_result_animation_impl: optimizer_folder=", optimizer_folder)
    print("show_result_animation_impl: joblist_folder=", joblist_folder)

    nodes = []
    for i, node in enumerate(os.listdir(joblist_folder)):
        # print([i], node)
        nodes.append(node)

    def plot_a_job(anim_ax, axis_info, node, joblist_folder, title_text):
        fig, axes = axis_info
        jobpath = os.path.join(joblist_folder, node)
        set_setting('optworking_folder', jobpath)
        result = selector_dialog.get_result(folder=jobpath)
        optimizer = result.get_optimizer()
        ret_ims = []
        for i, params in result.get_result_iterator():
            for ax in axes:
                ax.cla()
            axes[-1].grid(False)
            title_text.set_text("Result Animation of %s/%s-%03d" % (joblist_folder, node, i))
            optimizer.objective_func(params, plot=True, axis_info=axis_info)
            fig.tight_layout()
            fig.canvas.draw()
            buf = fig.canvas.buffer_rgba()
            im = anim_ax.imshow(buf)
            ret_ims.append([im])
        return ret_ims

    with plt.Dp(window_title="Result Animation", button_spec=["OK", "Cancel"]):
        anim_fig, anim_ax = plt.subplots(figsize=(18,5))
        anim_ax.set_axis_off()

        with plt.Dp():
            fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18,5))
            title_text = fig.suptitle("Result Animation of %s" % joblist_folder, fontsize=20)
            axt = ax2.twinx()
            axis_info= (fig, (ax1, ax2, ax3, axt))

            images = []
            for node in nodes:
                ims = plot_a_job(anim_ax, axis_info, node, joblist_folder, title_text)
                images += ims

        anim = ArtistAnimation(anim_fig, images, interval=200, blit=True,
                                repeat_delay=1000)
    
        anim.save(os.path.join(optimizer_folder, "result_animation.gif"))
        plt.show()


"""
    DecompDummyDialog.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""

from molass_legacy.KekLib.OurTkinter import Tk, Dialog

class DecompDummyDialog(Dialog):
    def __init__(self, parent, dialog, model):
        self.dialog = dialog
        self.model = model
        Dialog.__init__(self, parent, "DecompDummyDialog", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        from .DecompEditorFrame import DecompEditorFrame
        dialog = self.dialog
        frame = DecompEditorFrame(body_frame, dialog, dialog.sd, dialog.mapper, dialog.corbase_info, self.model)
        frame.pack()

def decompose(caller):
    dialog = caller.dialog
    model = dialog.get_current_frame().model
    dummy_dialog = DecompDummyDialog(dialog.parent, dialog, model)
    dummy_dialog.show()



"""
    FlowChange.py

    Copyright (c) 2021-2025, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from scipy.stats import linregress
from molass_legacy.KekLib.SciPyCookbook import smooth
from molass_legacy.KekLib.ExceptionTracebacker import log_exception
from .PeakRegion import PeakRegion
from .UvCorrectorEgh import UvCorrectorEgh
from .FlowChangeCandidates import get_largest_gradients
from .FlowChangePrep import get_easier_curve_y
from .FlowChangeUtils import is_ignorable_L_ratio, fit_sigmoid_impl
from .Sigmoid import ex_sigmoid, ex_sigmoid_inv

K_LIMIT_SAFE = 0.2
K_LIMIT_RISKY = 0.11            # < 0.117 for 20190305_1, > 0.108 for 20190227
K_LIMIT_GUARD_RATIO = 5         # < 5.92 for 20190305_1, > 3.32 for 20160227
CUSTOM_K_LIMIT = 0.95           # > 0.729 for OA, < 0.987 for HIF, 0.902 for Factin
K_LIMIT_SPECIAL = 0.05
L_RATIO_CUSTOM_LIMIT = 0.9      # < 0.996 for 20170309
R_VALUE_LIMIT = 0.9             # < 0.945 for SUB_TRN1, 
SAFE_RATIO = 0.99
SLICE_STOP_ALLOW = 10
MIN_DIST_ALLOW = 10
OUTSTANDING_RATIO_LIMIT = 5
SUPERSEDE_RATIO_ALLOW = 0.1
NARROW_MARGIN_RATIO = 0.05
SHOW_SPECIAL_CASE_MESSAGE = False

class FlowChange:
    def __init__(self, a_curve, a_curve2, x_curve, debug=False, fig_file=None):
        self.logger = logging.getLogger(__name__)
        self.a_curve = a_curve
        self.a_curve2 = a_curve2
        self.peak_region = PeakRegion(x_curve, a_curve, a_curve2)
        self.special_case_warned = False
        self.maybe_special = False

        def try_fc_model_fit(x, y, height, peak_region, std_p, pp3, slice_, ratio, outstanding_ratio, use_custom_sigmoid=True):
            print("============================= try_fc_model_fit: slice=", slice_)
            special = False
            failed = False
            try:
                x_ = x[slice_]
                y_ = y[slice_]
                assert len(x_) > 0
                ppk = [p for p in pp3 if x_[0] <= p and p <= x_[-1]]
                assert len(ppk) > 0
                popt, error_ratio = fit_sigmoid_impl(x_, y_, height, peak_region, std_p, ppk, outstanding_ratio, self.logger, debug=debug)
                if popt is None:
                    failed = True
                else:
                    if False:
                        debug_info = x_, y_
                    else:
                        debug_info = None
                    x1, y1, popt, special, L_ratio = self.get_safer_point(x, y, height, popt, ratio, debug_info, use_custom_sigmoid=use_custom_sigmoid)
            except:
                log_exception(self.logger, "fit_sigmoid error: ", n=10)
                failed = True

            if failed:
                x1, y1, popt, error_ratio, L_ratio = None, None, None, np.inf, 0

            return x1, y1, popt, error_ratio, L_ratio, special

        x = a_curve2.x
        y = a_curve2.y
        y_modified, y_for_gy, first_peak, last_peak = get_easier_curve_y(a_curve, a_curve2, self.peak_region, self.logger)

        uv_corrector = UvCorrectorEgh(a_curve, a_curve2)
        gy, pp3, outstanding_ratio, smoothed, corrected_y = get_largest_gradients(y_for_gy, 3, self.peak_region,
                                                return_full_info=True, uv_corrector=uv_corrector, logger=self.logger, debug=debug)

        self.gy = gy
        self.pp3 = pp3
        self.outstanding_ratio = outstanding_ratio
        if corrected_y is None:
            y_for_height = y_for_gy
        else:
            y_for_height = corrected_y

        height = np.max(y_for_height) - np.min(y_for_height)
        std_p = np.std(pp3)

        fc_xy = []
        popts = []
        error_ratios = []
        special_flags = []
        x0 = pp3[0]
        stop1 = first_peak[0]
        if abs(stop1 - x0) < SLICE_STOP_ALLOW:
            # as in 20170301
            stop1 = x0 + SLICE_STOP_ALLOW
        slices = [slice(0, stop1), slice(last_peak[-1], None), slice(0, None)]
        num_tests = 0
        num_candidates = 0
        for slice_, ratio in zip(slices, [SAFE_RATIO, 1 - SAFE_RATIO, SAFE_RATIO]):
            x1, y1, popt, error_ratio, L_ratio, special = try_fc_model_fit(x, y_for_gy, height, self.peak_region, std_p, pp3, slice_, ratio, outstanding_ratio)
            if special:
                self.maybe_special = True
            # print([num_tests], "-------------- x1, y1, popt, error_ratio=", x1, y1, popt, error_ratio)

            num_tests += 1
            if x1 is not None:
                num_candidates += 1

            if num_tests == 3:
                if x1 is not None:
                    if x1 > last_peak[1]:
                        # as in 20200630_3 which should be investigated.
                        self.logger.info("third trial at %.3g in the descending side of the last peak is discarded.", x1)
                    else:
                        # as in pH6
                        fc_xy[0] = (x1, y1, L_ratio, special)    # although it will finally be dismissed
                        popts[0] = popt
            else:
                fc_xy.append((x1, y1, L_ratio, special))
                popts.append(popt)
                error_ratios.append(error_ratio)
                special_flags.append(special)

            if num_tests == 2 and num_candidates > 0:
                break

        draw_debug_plot = debug or fig_file is not None
        if draw_debug_plot:
            import molass_legacy.KekLib.DebugPlot as plt
            def fc_result_plot(gy, pp3, title_trailer=""):
                from matplotlib.patches import Rectangle
                from molass_legacy._MOLASS.SerialSettings import get_setting
                from DataUtils import get_in_folder
                from molass_legacy.Elution.CurveUtils import simple_plot
                from molass_legacy.UV.PlainCurveUtils import get_flat_wavelength

                plt.push()
                fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18,6))

                in_folder = get_in_folder()
                fig.suptitle("Flowchange Recognition on %s%s" % (in_folder, title_trailer), fontsize=20)
                lambda_list = [get_setting("absorbance_picking"), get_flat_wavelength()]
                ax1.set_title("Elution Curve (1) at λ=%.3g" % (lambda_list[0]), fontsize=16)
                ax2.set_title("Elution Curve (2) at λ=%.3g" % (lambda_list[1]), fontsize=16)
                ax3.set_title("Gradient of the Curve (2)", fontsize=16)

                simple_plot(ax1, a_curve, legend=False)
                ymin1, ymax1 = ax1.get_ylim()
                ax1.set_ylim(ymin1, ymax1)

                f, t = self.peak_region.get_wider_ends()
                p = Rectangle(
                        (f, ymin1),     # (x,y)
                        t - f,          # width
                        ymax1 - ymin1,  # height
                        facecolor   = 'cyan',
                        alpha       = 0.1,
                        label = "wider peak region",
                    )
                ax1.add_patch(p)
                ax1.legend()

                ax2.plot(x, y)
                if y_for_gy is not None:
                    ax2.plot(x, y_for_gy, ':', color="gray")
                if corrected_y is not None:
                    ax2.plot(x, corrected_y, color="green")
                ymin2, ymax2 = ax2.get_ylim()
                ax2.set_ylim(ymin2, ymax2)

                color = "green" if smoothed else None
                ax3.plot(x, gy, color=color)
                for i, p in enumerate(pp3):
                    ax3.plot(p, gy[p], "o", label="%d-th significant" % i)

                xmin3, xmax3 = ax3.get_xlim()
                ymin3, ymax3 = ax3.get_ylim()
                tx = (xmin3 + xmax3)/2
                if abs(ymin3) < abs(ymax3):
                    ty = ymax3/2
                else:
                    ty = ymin3/2
                smoothed_str = "\n(of the smoothed)" if smoothed else ""
                ax3.text(tx, ty, "outstanding_ratio=%.3g%s" % (outstanding_ratio, smoothed_str), fontsize=20, alpha=0.5, ha="center")
                ax3.legend()

                for popt, (x1, y1, L_ratio, special), slice_ in zip(popts, fc_xy, slices):
                    if x1 is None:
                        pass
                    else:
                        if self.peak_region.is_in_the_region(x1):
                            color0 = "gray"
                            color1 = "gray"
                            color1p = "gray"
                        elif special:
                            color0 = "cyan"
                            color1 = "cyan"
                            color1p = "cyan"
                        else:
                            color0 = "red"
                            color1 = "yellow"
                            color1p = "orange"

                        x_ = x[slice_]
                        y_ = y[slice_]
                        L, x0, k, b, s1, s2 = popt
                        ax2.plot(x_, ex_sigmoid(x_, *popt))
                        ax2.plot([x0, x0], [ymin2, ymax2], ":", color=color0)
                        ax2.plot([x1, x1], [ymin2, ymax2], color=color1)
                        ax2.plot(x1, y1, "o", color=color1p)
                        xmin2, xmax2 = ax2.get_xlim()
                        tx = x1
                        w = 0.5
                        ty = ymin2 * (1-w) + ymax2 * w
                        ax2.text(tx, ty, "L_ratio=%.3g, k=%.3g" % (L_ratio, k), fontsize=20, alpha=0.5, ha="center")
                        ax1.plot([x1, x1], [ymin1, ymax1], color=color1)
                        ax1.plot(x1, a_curve.spline(x1), "o", color=color1p)

                fig.tight_layout()
                fig.subplots_adjust(top=0.85)

                if fig_file is None:
                    plt.show()
                else:
                    from time import sleep
                    plt.show(block=False)
                    fig = plt.gcf()
                    fig.savefig(fig_file)
                    sleep(1)
                plt.pop()

            fc_result_plot(gy, pp3)

        has_special_ = False
        for n in range(2):
            x1, y1, L_ratio, special = fc_xy[n]
            if x1 is None:
                continue

            if special:
                has_special_ = True

            if self.peak_region.is_in_the_region(x1):
                # as in pH6
                # although similar case as in 20161202 does not fall herein
                fc_xy[n] = None, None, 0, False
                self.logger.info("this flow change at %.4g included in the peak %s has been dismissed.", x1, str(self.peak_region.get_ends()))
            else:
                popt = popts[0]
                if popt is None:
                    # as in 20191006_proteins5
                    continue

                x0 = popt[1]
                pp3d = np.abs(pp3 - x0)
                min_dist = np.min(pp3d)
                max_dist = np.max(pp3d)
                dist_ratio = self.peak_region.get_distance_ratio(x0)
                print("dist_ratio=%.3g" % dist_ratio)
                print("outstanding_ratio=%.3g, (min_dist, max_dist)=(%.3g, %.3g)" % (outstanding_ratio, min_dist, max_dist))
                outstanding = outstanding_ratio > OUTSTANDING_RATIO_LIMIT
                if (max_dist > MIN_DIST_ALLOW
                    and min(*pp3[1:3]) < first_peak[0]      # pp3[1] < first_peak[0] for OAGIwyatt_02
                    ):
                    error_ratio = error_ratios[0]
                    j = int(round(x1))
                    if j < first_peak[1]:
                        slice_ = slice(j, first_peak[1])
                        if min_dist < MIN_DIST_ALLOW:
                            # as in OAGIwyatt_02, 20161202
                            x1_, y1_, popt_, error_ratio_, L_ratio_, special_ = try_fc_model_fit(x, y_for_gy, height, self.peak_region, std_p, pp3, slice_,
                                                                                                    SAFE_RATIO, outstanding_ratio)
                            fix_type = "1"
                            gy_for_debug = gy
                            pp3_for_debug = pp3
                        else:
                            # as in 20170304
                            sy = smooth(y_for_gy)
                            sgy, pp3_ = get_largest_gradients(sy, 3, self.peak_region)
                            std_p_ = np.std(pp3_)
                            x1_, y1_, popt_, error_ratio_, L_ratio_, special_ = try_fc_model_fit(x, sy, height, self.peak_region, std_p_, pp3_, slice_,
                                                                                                    SAFE_RATIO, outstanding_ratio, use_custom_sigmoid=False)
                            fix_type = "2"
                            gy_for_debug = sgy
                            pp3_for_debug = pp3_
                        print("(error_ratio_, error_ratio+allow)=(%.3g, %.3g), special=%s" % (error_ratio_, error_ratio + SUPERSEDE_RATIO_ALLOW, str(special_)) )
                        if (x1_ is not None
                            and not self.peak_region.is_in_the_region(x1_)
                            and error_ratio_ < error_ratio + SUPERSEDE_RATIO_ALLOW
                            ):
                            # as in OAGIwyatt_02 (but not in 20161202)
                            # how abount special cases?
                            fc_xy[n] = x1_, y1_, L_ratio_, special_
                            popts[n] = popt_
                            if special_:
                                self.maybe_special = True

                            self.logger.info("next flow change[%d] at %.4g has been superseded with fix_type=%s and error_ratio %.3g < %.3g + %.3g", n, x1_, fix_type, error_ratio_, error_ratio, SUPERSEDE_RATIO_ALLOW)
                            if draw_debug_plot:
                                fc_result_plot(gy_for_debug, pp3_for_debug, title_trailer=" (after fix%s)" % fix_type)
                    else:
                        # 
                        pass

        self.has_special_ = has_special_
        self.fcj = [None if x1 is None or special else int(round(x1)) for x1, y1, L_ratio, special in fc_xy]
        self.popts = popts

    def has_special(self):
        return self.has_special_

    def get_safer_point(self, x, y, height, popt, ratio, debug_info, use_custom_sigmoid=True):

        special = False
        L, x0, k, b, s1, s2 = popt
        L_ratio = abs(L)/height

        if debug_info is not None:
            import molass_legacy.KekLib.DebugPlot as plt
            print("get_safer_point: L=%.3g, L_ratio=%.3g, k=%.3g" % (L, L_ratio, k))
            x_, y_ = debug_info
            plt.push()
            fig, ax = plt.subplots()
            ax.set_title("get_safer_point debug (1)")
            ax.plot(x, y)
            ax.plot(x_, y_, ":")
            ax.plot(x, ex_sigmoid(x, *popt))
            plt.show()
            plt.pop()

        if is_ignorable_L_ratio(L_ratio):
            self.logger.info("a flowchange candidate at %.3g ignored with L_ratio=%.3g", x0, L_ratio)
            return None, None, popt, special, L_ratio

        if (k > K_LIMIT_SAFE
            or k > K_LIMIT_RISKY and self.outstanding_ratio > K_LIMIT_GUARD_RATIO
            ):
            print("k=%.3g > K_LIMIT, L_ratio=%.3g" % (k, L_ratio))
            if use_custom_sigmoid and k < CUSTOM_K_LIMIT or L_ratio > L_RATIO_CUSTOM_LIMIT:
                from .CustomSigmoid import get_safer_point_impl
                try:
                    x1, y1 = get_safer_point_impl(x, y, ratio, popt, debug=False)
                    use_sigmoid = False
                except AssertionError:
                    # as in Matsumura
                    log_exception(self.logger, "get_safer_point_impl failure: ")
                    use_sigmoid = True
            else:
                use_sigmoid = True

            if use_sigmoid:
                y1 = b + L * ratio
                x1 = ex_sigmoid_inv(y1, L, x0, k, b, s1, s2)
                # x1 can exceed limits as in 20160227
                # print("--------------------- x[0], x1, x[-1]=", x[0], x1, x[-1])

        elif k > K_LIMIT_SPECIAL:
            r_value = linregress(x, y)[2]
            print("r_value=", r_value)
            if r_value > R_VALUE_LIMIT:
                # as in SUB_TRN1
                # getting a safer point is not appropriate in this case
                self.logger.info("this flowchange candidate with k=%.3g is a special case due to r_value=%.3g", k, r_value)
                x1, y1 = x0, ex_sigmoid(x0, *popt)
            else:
                x1, y1 = None, None
        else:
            x1, y1 = None, None
        if x1 is None:
            self.logger.info("a flowchange candidate ignored with k=%.3g", k)

        if x1 is not None:
            assert x[0] <= x1 and x1 <= x[-1]
            special, ratio = self.peak_region.is_special(x1, return_ratio=True)
            if special and SHOW_SPECIAL_CASE_MESSAGE:
                from molass_legacy._MOLASS.SerialSettings import get_setting
                test_pattern = get_setting("test_pattern")
                if test_pattern is None and not self.special_case_warned:
                    try:
                        import molass_legacy.KekLib.CustomMessageBox as MessageBox
                        from molass_legacy._MOLASS.Version import get_version_string
                        parent = plt.get_parent()       # note that enable_debug_plot==1 and this parent it the main dialog
                        if get_version_string().find("_MOLASS 1") >= 0:
                            tool_name = "_MOLASS ver. 2. or later"
                        else:
                            tool_name = '"Full Optimization"'
                        MessageBox.showwarning("Special Flowchange Warning",
                            "A special flow change(*) has been detected in this data set.\n"
                            "Analysis using %s is recommended.\n"
                            'Otherwise, trim manually at "Data Rage".\n\n'
                            "* this point will not appear in figures in this version." % tool_name,
                            parent=parent)
                    except:
                        # ctypes.ArgumentError: argument 1: <class 'OverflowError'>: int too long to convert
                        log_exception(self.logger, "Special Flowchange Warning failure: ")
                    self.special_case_warned = True
                self.logger.info("this flowchange candidate at %.4g is discarded as a special case with ratio=%.3g.", x1, ratio)
                # how about error_ratio in this case

        return x1, y1, popt, special, L_ratio

    def get_raw_flow_changes(self):
        return self.fcj

    def get_flow_changes(self, narrow=False):
        ret_list = []
        for k, fc in enumerate(self.fcj):
            if fc is None:
                fc = 0 if k == 0 else len(self.a_curve2.x) - 1
            else:
                if narrow:
                    margin = int(self.peak_region.get_size()*NARROW_MARGIN_RATIO)   # get_size returns size with UV elution units
                    sign = +1 if k == 0 else -1
                    fc += sign*margin
            ret_list.append(fc)
        return ret_list

    def get_safe_flow_changes(self):
        # used in 'narrow_trimming'
        return self.get_flow_changes(narrow=True)

    def get_real_flow_changes(self):
        # consider unifying with get_raw_flow_changes using 20211021
        flow_changes_src = self.get_flow_changes()
        flow_changes = []
        for fc in flow_changes_src:
            if fc == 0 or fc == len(self.a_curve2.x) - 1:
                fc_ = None
            else:
                fc_ = fc
            flow_changes.append(fc_)
        return flow_changes        # print( 'right_jump_j=', self.right_jump_j )

    def get_mapped_flow_changes(self):
        similarity = self.get_similarity()
        flow_changes = self.get_real_flow_changes()
        return [j if j is None else similarity.inverse_int_value(j) for j in flow_changes]

    def get_similarity(self):
        return self.peak_region.similarity

    def get_solver_info(self, debug=False):
        if debug:
            from importlib import reload
            import Trimming.UvBaseSolverInfo
            reload(Trimming.UvBaseSolverInfo)
        from .UvBaseSolverInfo import make_info_lists
        return make_info_lists(self, debug=debug)

    def remove_irregular_points(self):
        ret_pp = []
        err_pp = []
        size = len(self.a_curve2.y)
        for k in self.pp3:
            if k < size - 10:
                ret_pp.append(k)
            else:
                err_pp.append(k)
        if len(err_pp) > 0:
            ret_slice = slice(None, np.min(err_pp))
        else:
            ret_slice = slice(None, None)
        return np.array(sorted(ret_pp, key=lambda j: -abs(self.gy[j]))), np.array(err_pp), ret_slice



"""
    Optimizer.ModifyStateDialog.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, root
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar

class ModifyStateDialog(Dialog):
    def __init__(self, parent, dialog):
        self.dialog = dialog
        self.current_optimizer = dialog.fullopt
        self.orig_params = dialog.canvas.get_current_params()
        self.current_params = self.orig_params
        self.dsets = dialog.canvas.dsets
        self.applied = False
        Dialog.__init__(self, parent, "Modify State", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()
        bframe = Tk.Frame(body_frame)
        bframe.pack(fill=Tk.X, padx=20)

        tframe = Tk.Frame(bframe)
        tframe.pack(side=Tk.LEFT)
        pframe = Tk.Frame(bframe, width=100)
        pframe.pack(side=Tk.RIGHT)

        fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18, 5))
        axt = ax2.twinx()
        self.fig = fig
        self.axes = (ax1, ax2, ax3, axt)

        axt.grid(False)
        axis_info = fig, self.axes

        fv = 0
        ax1.set_title("UV Decomposition", fontsize=16)
        ax2.set_title("Xray Decomposition", fontsize=16)
        ax3.set_title("Objective Function Scores in SV=%.3g" % fv, fontsize=16)

        self.current_optimizer.objective_func(self.orig_params, plot=True, axis_info=axis_info)

        fig.tight_layout()
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)

        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()

        button = Tk.Button(pframe, text="Show Parameters", command=self.show_parameters)
        button.pack(side=Tk.RIGHT, padx=20)

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=20, pady=5)
        w = Tk.Button(box, text="Evaluate a new suggestion", width=30, command=self.evaluate)
        w.pack(side=Tk.LEFT, padx=20, pady=5)
        w = Tk.Button(box, text="Start a new optimization process", width=30, command=self.start)
        w.pack(side=Tk.LEFT, padx=20, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def apply(self):
        from .InitialInfo import PeProxy

        print("apply")
        self.applied = True
        pdialog = self.dialog
        optinit_info = pdialog.optinit_info
        self.pe_proxy = PeProxy(optinit_info, self.current_optimizer, self.current_params)

    def get_pe_proxy(self):
        return self.pe_proxy

    def show_parameters(self, debug=True):
        print("show_parameters")
        if debug:
            from importlib import reload
            import Optimizer.ParamsInspection
            reload(Optimizer.ParamsInspection)
        from .ParamsInspection import ParamsInspection
        # self.dialog.grab_set()  # temporary fix to the grab_release problem
        parent = self.dialog.parent
        dialog = ParamsInspection(parent, self.current_params, self.dsets, self.current_optimizer)
        dialog.show()
        # self.dialog.grab_set()  # temporary fix to the grab_release problem

    def evaluate(self):
        print("evaluate")
        temp_optimizer, new_params = self.make_a_new_suggestion()
        self.current_optimizer = temp_optimizer
        self.current_params = new_params

        fv = 0

        for ax in self.axes:
            ax.cla()
        ax1, ax2, ax3, axt = self.axes
        axt.grid(False)
        ax1.set_title("Evaluated UV Decomposition", fontsize=16)
        ax2.set_title("Evaluated Xray Decomposition", fontsize=16)
        ax3.set_title("Evaluated Objective Function Scores in SV=%.3g" % fv, fontsize=16)
        axis_info = self.fig, (ax1, ax2, ax3, axt)
        temp_optimizer.objective_func(self.current_params, debug=True, axis_info=axis_info)
        self.mpl_canvas.draw()

    def make_a_new_suggestion(self, debug=True):
        if debug:
            from importlib import reload
            import Optimizer.FunctionDebuggerUtils
            reload(Optimizer.FunctionDebuggerUtils)
        from .FunctionDebuggerUtils import create_optimizer_for_debug

        js_canvas = self.dialog.canvas
        n_components = js_canvas.fullopt.n_components
        init_params = js_canvas.optinit_info.init_params

        params = init_params

        separate_params = js_canvas.fullopt.split_params_simple(params)
        n, new_params, new_composite = self.get_enhanced_params(n_components, separate_params)

        temp_optimizer = create_optimizer_for_debug(
                    self.dsets, n, js_canvas.optinit_info,
                    None,                                   # not used when prepare=False
                    composite=new_composite, prepare=False)

        temp_optimizer.prepare_for_optimization(new_params)

        return temp_optimizer, new_params

    def get_enhanced_params(self, n, separate_params):
        from .CompositeInfo import CompositeInfo

        xr_params, xr_baseparams, rg_params, mapping, uv_params = separate_params[0:5]

        ne = xr_params.shape[0]

        if n == 6:
            new_xr_params = []
            new_rg_params = []
            new_uv_params = []
            for i in range(ne):
                xr = xr_params[i].copy()
                rg = rg_params[i]
                uv = uv_params[i]

                if i == 1:
                    xr[0] *= 0.5
                    uv *= 0.5
                    new_xr_params.append(xr)
                    new_rg_params.append(rg)
                    new_uv_params.append(uv)

                new_xr_params.append(xr)
                new_rg_params.append(rg)
                new_uv_params.append(uv)

            new_xr_params = np.array(new_xr_params)
            new_rg_params = np.array(new_rg_params)
            new_uv_params = np.array(new_uv_params)

            new_params = np.concatenate([new_xr_params.flatten(), xr_baseparams, new_rg_params, mapping, new_uv_params] + separate_params[5:])
            return n + 1, new_params, CompositeInfo(composites=[[0], [1,2], [3,4], [5], [6]])

        elif n == 7:
            new_params = np.concatenate([xr_params.flatten()] + separate_params[1:])
            return n, new_params, CompositeInfo(composites=[[0,1,2], [3,4], [5], [6]])

        else:
            # i.e., no change
            new_params = np.concatenate([xr_params.flatten()] + separate_params[1:])
            return n, new_params, self.current_optimizer.composite

    def start(self):
        import molass_legacy.KekLib.CustomMessageBox as MessageBox

        reply = MessageBox.askokcancel("Restart Confirmation",
                    "This action will create a new analysis folder\n"
                    "and start a new optimization process with\n"
                    "the modified parameters.\n"
                    "Are you sure to continue?",
                    parent=self,
                    )
        if reply:
            self.ok()
        else:
            self.cancel()



# coding: utf-8
"""
    DecompInfo.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import copy

class DecompInfo:
    def __init__(self, opt_recs, opt_recs_uv, fx):
        self.opt_recs = opt_recs
        self.opt_recs_uv = opt_recs_uv
        self.fx = fx

    def get_scaled_recs(self, conc_factor):
        ret_recs = []
        for rec_xr, rec_uv in zip(self.opt_recs, self.opt_recs_uv):
            rec = rec_xr.scale_to(rec_uv, conc_factor)
            ret_recs.append(rec)
        return ret_recs



"""
    GpuInfo.py

    Copyright (c) 2019-2024, SAXS Team, KEK-PF
"""
import sys
import os
import subprocess
import re

class DxdiagInfo:
    def __init__(self):
        import os
        import tempfile
        import subprocess

        self.info_list = []
        path = tempfile.gettempdir()
        # print(path)
        file = path + '/dxdiag.txt'
        ret = subprocess.run(['dxdiag', '/t', file])
        if ret.returncode == 0:
            with open(file) as fh:
                for line in fh.readlines():
                    # print(line)
                    kw = 'Card name:'
                    n = line.find(kw)
                    if n > 0:
                        n += len(kw) + 1
                        self.cardname = line[n:-1]
                        # print(self.cardname)
                        self.info_list.append(line[:-1])
                        break
        else:
            assert False

        os.remove(file)

    def get_info_list(self):
        return self.info_list

def get_cupy_version():
    CUDA_PATH = os.environ.get('CUDA_PATH')
    if CUDA_PATH is None:
        cupy_cuda = 'cupy-cuda11x'
    else:
        cuda_ver = CUDA_PATH.split('\\')[-1]
        # print('cuda_ver=', cuda_ver)
        # cuda_ver is like v11.8, v12.3, etc.
        cupy_cuda = 'cupy-cuda%sx' % cuda_ver[1:3]
    try:
        result = subprocess.run([sys.executable, '-m', 'pip', 'show', cupy_cuda], capture_output=True)
        assert result.returncode == 0
        stdout = result.stdout.decode()
        version_re = re.compile(r'Version: (\S+)\r\n')
        m = version_re.search(stdout)
        if m:
            version = (cupy_cuda, m.group(1))
        else:
            version = None
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(None, "get_cupy_version falure: ")
        version = None
    # print("get_cupy_version: version=", version)
    return version

def get_nvcc_version():
    try:
        result = subprocess.run(['nvcc', '-V'], capture_output=True)
        stdout = result.stdout.decode()
        # print('get_nvcc_version: stdout=', stdout)
        version_re = re.compile('(V\d+\.\d+)\.\d+')
        m = version_re.search(stdout)
        if m:
            version = m.group(1)
        else:
            version = version
    except:
        version = None
    # print('get_nvcc_version: version=', version)
    return version

def get_available_cuda_versions():
    try:
        result = subprocess.run(['where', 'nvcc'], capture_output=True)
        stdout = result.stdout.decode()
    except:
        stdout = ""

    versions = []
    v11x = []
    for line in stdout.split("\n"):
        nodes = line.split("\\")
        if len(nodes) > 2:
            versions.append(nodes[-3])

    return versions + v11x

class GpuInfo:
    def __init__(self):
        self.get_graphics_card_info()
        self.cuda_path = os.environ.get('CUDA_PATH')
        self.cupy_version = get_cupy_version()
        self.nvcc_version = get_nvcc_version()
        self.cuda_versions = get_available_cuda_versions()
        self.cuda_version = self.cuda_versions[0] if len(self.cuda_versions) > 0 else None
        self.cuda_is_availale = self.cuda_version is not None
        self._cupy_ok = None
        try:
            import cupy
            self.cupy_is_installed = True
        except:
            self.cupy_is_installed = False

    def get_graphics_card_info(self):
        from pythoncom import CoInitialize, CoUninitialize
        CoInitialize()      # required inside a thread
        import wmi
        wmi_infos = wmi.WMI().Win32_VideoController()

        # Name = "NVIDIA GeForce GTX 960";
        vp_re = re.compile(r'\s+Name = "(.+)"')
        self.card_infos = []

        exists_compatible = False
        self.index_compatible = None
        for k, i in enumerate(wmi_infos):
            info = str(i)
            m = vp_re.search(info)
            if m:
                name = m.group(1)
            else:
                name = None

            if info.find('NVIDIA') > 0:
                compatible = True
                exists_compatible = True
                if self.index_compatible is None:
                    # set the first-found
                    self.index_compatible = k
            else:
                compatible = False

            self.card_infos.append((name, compatible))

        self.nvidia_compatible = exists_compatible
        k = 0 if self.index_compatible is None else self.index_compatible
        self.card_name = self.card_infos[k][0]
        CoUninitialize()    # required inside a thread

    def check_cupy_cuda_consistency(self):
        cupy_ver_num = self.cupy_version[1].replace('cuda', '')
        non_digit_re = re.compile(r'\D+')
        cuda_ver_num = re.sub(non_digit_re, '', self.nvcc_version)
        # print('cupy_ver_num=', cupy_ver_num, 'cuda_ver_num', cuda_ver_num)
        return cupy_ver_num[0:2] == cuda_ver_num[0:2]   # "12x"[0:2] == "123"[0:2]

    def cuda_ok(self):
        return self.nvidia_compatible and self.cuda_is_availale

    def cupy_ok(self):
        if self._cupy_ok is not None:
            return self._cupy_ok

        if self.cuda_ok():
            if not self.cupy_is_installed:
                self.cupy_ng_reason = "However, no consistent cupy is installed."
                self._cupy_ok = False
                return False
        else:
            self.cupy_ng_reason = "However, cuda is not available."
            self._cupy_ok = False
            return False

        ok = False
        if self.cupy_version is None:
            cupy_cuda_version = "v11.x"
        else:
            cupy_cuda_version = re.sub(r"cupy-cuda(\d{2})(\w)",  lambda m: "v%s.%s" % (m.group(1), m.group(2)), self.cupy_version[0])
            # print("cupy_cuda_version=", cupy_cuda_version)
            for k, v in enumerate(self.cuda_versions):
                # print([k], v)
                if cupy_cuda_version[0:3] == v[0:3]:    # eg., "12.x"[0:3] == "12.3"[0:3]
                    ok = True
                    break
        if ok:
            self.cupy_ng_reason = None
        else:
            self.cupy_ng_reason = "However, cupy cuda version %s is not installed" % cupy_cuda_version

        self._cupy_ok = ok
        return ok

    def get_reason_texts(self):
        texts = []

        # GPU Card compatibility
        not_ = '' if self.nvidia_compatible else 'not '
        texts.append(self.card_name + ' is %sNVIDIA compatible.' % not_)
        if not self.nvidia_compatible:
            # unless it is nvidia_compatible, there is no need to add further information.
            return texts

        # CUDA installation
        conn_text = 'And, ' if self.cuda_is_availale == self.nvidia_compatible else 'However, '
        if self.cuda_version is None:
            if self.nvcc_version is None:
                cuda_text = 'CUDA is not found installed.'
            else:
                versions = (self.cupy_version, self.nvcc_version)
                if self.check_cupy_cuda_consistency():
                    cuda_text = 'Something is wrong with Cupy %s and CUDA %s.' % versions
                else:
                    cuda_text = 'Cupy %s is inconsistent with CUDA %s.' % versions
        else:
            be = "is" if len(self.cuda_versions) == 1 else "are"
            cuda_text = 'CUDA %s %s installed.' % (",".join(self.cuda_versions), be)
        texts.append(conn_text + cuda_text)

        # CUPY avaiability
        if not self.cupy_ok():
            texts.append(self.cupy_ng_reason)

        return texts



"""
    DraggableListbox.py

    adapted from
        https://stackoverflow.com/questions/14459993/tkinter-listbox-drag-and-drop-with-python

"""

import tkinter as tk

class DraggableListbox(tk.Listbox):
    """ A tk listbox with drag'n'drop reordering of entries. """
    def __init__(self, master, **kw):
        kw['selectmode'] = tk.MULTIPLE
        kw['activestyle'] = 'none'
        tk.Listbox.__init__(self, master, kw)
        self.bind('<Button-1>', self.getState, add='+')
        self.bind('<Button-1>', self.setCurrent, add='+')
        self.bind('<B1-Motion>', self.shiftSelection)
        self.curIndex = None
        self.curState = None
    def setCurrent(self, event):
        ''' gets the current index of the clicked item in the listbox '''
        self.curIndex = self.nearest(event.y)
    def getState(self, event):
        ''' checks if the clicked item in listbox is selected '''
        i = self.nearest(event.y)
        self.curState = self.selection_includes(i)
    def shiftSelection(self, event):
        ''' shifts item up or down in listbox '''
        i = self.nearest(event.y)
        if self.curState == 1:
            self.selection_set(self.curIndex)
        else:
            self.selection_clear(self.curIndex)
        if i < self.curIndex:
            # Moves up
            x = self.get(i)
            selected = self.selection_includes(i)
            self.delete(i)
            self.insert(i+1, x)
            if selected:
                self.selection_set(i+1)
            self.curIndex = i
        elif i > self.curIndex:
            # Moves down
            x = self.get(i)
            selected = self.selection_includes(i)
            self.delete(i)
            self.insert(i-1, x)
            if selected:
                self.selection_set(i-1)
            self.curIndex = i

if __name__ == '__main__':
    root = tk.Tk()
    listbox = DraggableListbox(root)
    for i,name in enumerate(['name'+str(i) for i in range(10)]):
        listbox.insert(tk.END, name)

        if False:
            if i % 2 == 0:
                listbox.selection_set(i)
    listbox.pack(fill=tk.BOTH, expand=True)
    root.mainloop()



"""
    Optimizer.OptimizerMain.py

    Copyright (c) 2021-2024, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.Baseline.BaselineUtils import create_xr_baseline_object
from .FuncImporter import import_objective_function

def optimizer_main(in_folder, trimming_txt=None, n_components=3,
                   solver=None,
                   init_params=None, real_bounds=None,
                   drift_type=None, niter=100, seed=None,
                   callback=True, class_code='F0000', shared_memory=None,
                   nnn=0,  debug=True):
    from .FullOptInput import FullOptInput

    fullopt_input = FullOptInput(in_folder=in_folder, trimming_txt=trimming_txt)
    dsets = fullopt_input.get_dsets()

    if seed is None:
        seed = np.random.randint(100000, 999999)

    fullopt_class = import_objective_function(class_code)
    uv_base_curve = fullopt_input.get_base_curve()      # uv_base_curve comes from FullOptInput.get_sd_from_folder()
    xr_base_curve = create_xr_baseline_object()
    optimizer = fullopt_class(dsets, n_components,
                uv_base_curve=uv_base_curve,
                xr_base_curve=xr_base_curve,
                qvector=fullopt_input.sd.qvector,   # trimmmed sd
                wvector=fullopt_input.sd.lvector,
                shared_memory=shared_memory)

    strategy = optimizer.get_strategy()
    if strategy.trust_initial_baseline():
        baseline_fixed = True
        from molass_legacy.Optimizer.FixedBaselineOptimizer import FixedBaselineOptimizer
        fb_optimizer = FixedBaselineOptimizer(optimizer)
        result = fb_optimizer.solve(init_params, real_bounds=real_bounds, niter=niter, seed=seed, debug=debug)
    else:
        baseline_fixed = False
        if strategy.baseline_first():
            if nnn == 0:
                from molass_legacy.Optimizer.BaselineOptimizer import BaselineOptimizer
                baseline_optimizer = BaselineOptimizer(optimizer)
                baseline_indeces = optimizer.get_baseline_indeces()
                result = baseline_optimizer.solve(init_params, baseline_indeces)
            else:
                baseline_fixed = True
                result = optimizer.solve(init_params, real_bounds=real_bounds, niter=niter, seed=seed, callback=callback, method=solver,
                                        baseline_fixed=baseline_fixed, debug=debug)
        else:
            if strategy.is_strategic(nnn):
                from molass_legacy.Optimizer.StrategicOptimizer import StrategicOptimizer
                temp_params = init_params
                indeces_list = strategy.get_indeces_list(nnn)
                assert len(indeces_list) == 1   # for now
                open_mode = "w"                 # "callback.txt" open mode
                for indeces in indeces_list:
                    strategic_optimizer = StrategicOptimizer(optimizer, indeces)
                    # task: add method option
                    result = strategic_optimizer.solve(temp_params, real_bounds=real_bounds, niter=niter, seed=seed, open_mode=open_mode, debug=debug)
                    temp_params = result.x
                    open_mode = "a"
            else:
                result = optimizer.solve(init_params, real_bounds=real_bounds, niter=niter, seed=seed, callback=callback, method=solver, debug=debug)
    fig_info = [in_folder, None, result]
    if debug:
        optimizer.objective_func(result.x, plot=True, fig_info=fig_info)



"""
    DenssManagerDialog.py

    Copyright (c) 2019-2025, SAXS Team, KEK-PF
"""
import numpy as np
from time import strftime, localtime
import logging
import queue
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk
from molass_legacy.KekLib.TkUtils import get_widget_geometry
from molass_legacy.KekLib.TkSupplements import set_icon
from molass_legacy.KekLib.ScrolledFrame import ScrolledFrame
from .DenssUtils import MAXNUM_STEPS

dialogs = []
# DIALOG_HEIGHT = 250
PROGRES_BAR_SCALE = 3
PROGRES_BAR_WIDTH = 240 # 80*3
REFRESH_INTERVAL = 500
VALUE_FORMATS = ["%d", "%4.2e", "%3.2f", "%5i"]
STATE_TEXT = ["running", "started", "waiting", "finished", "error", "canceled"]
STATE_COLOR = ["black", "black", "black", "green", "red", "orange"]

def format_time(t):
    return "" if t is None else strftime("%H:%M:%S", localtime(t))

def format_duration(t):
    if t is None:
        return ""
    else:
        minutes, seconds = divmod(int(t), 60)
        return "%d:%02d" % (minutes, seconds) if minutes > 0 else "%d" % (seconds)

def get_state_text(state):
    i = abs(state)
    return STATE_TEXT[i]

def get_state_color(state):
    i = abs(state)
    return STATE_COLOR[i]

def get_ch2_fg_color(value):
    color = 'black'

    if value is not None:
        if value > 0.1:
            color = 'red'
        elif value > 0.01:
            color = 'orange'

    return color

def show_manager_dialog(parent, jobs=None):

    if len(dialogs) > 0:
        dialog = dialogs[0]
        dialog.focus_force()
        return

    from .DenssManager import activate_manager, get_list
    from molass_legacy.KekLib.BasicUtils import get_home_folder
    from molass_legacy.Env.EnvInfo import get_global_env_info

    log_folder = get_home_folder() + '/log'
    env_info = get_global_env_info(gpu_info=True)
    use_gpu = env_info.nvidiagpu_is_available
    num_workers = activate_manager(log_folder, use_gpu)

    dialog = DenssManagerDialog(parent, num_workers, use_gpu)
    dialogs.append(dialog)

    def post_processor():
        dialog.update()

        if jobs is not None:
            dialog.submit_jobs(jobs)

        job_list = get_list()
        num_jobs = max(1, len(job_list))
        dialog.update_geometry(num_jobs)

    parent.after(100, post_processor)

    dialog.show()
    dialogs.pop()

def terminate_manager(parent):
    from .DenssManager import get_list
    job_list = get_list()
    num_remaining_jobs = 0
    for i, row_rec in enumerate(job_list):
        job_id, job_rec = row_rec
        state = job_rec[2]
        if state > -3:
            num_remaining_jobs += 1
    if num_remaining_jobs > 0:
        import molass_legacy.KekLib.CustomMessageBox as MessageBox
        MessageBox.showinfo(
                'Cancel All Confirmation',
                'There are %d DENSS jobs remaininig.\n' % num_remaining_jobs
                + 'You must wait or cancel them before exit\n'
                + 'in the "DENSS Manager Dialog", which will appear next.',
                parent=parent,
                )
        show_manager_dialog(parent)
        return False

    from .DenssManager import terminate_manager_impl
    terminate_manager_impl()
    return True

class DenssManagerDialog(Dialog):
    def __init__(self, parent, num_workers, use_gpu):
        self.grab = 'local'     # used in grab_set
        self.parent = parent
        self.num_workers = num_workers
        self.use_gpu = use_gpu
        self.repeat_refresh = True
        self.basic_geometry = None
        Dialog.__init__(self, parent, "DENSS Process Manager", visible=False)

    def show(self):
        self.parent.after(REFRESH_INTERVAL, self.refresh)
        self._show()

    def body(self, body_frame):
        set_icon( self )

        t_frame = Tk.Frame(body_frame)
        t_frame.pack(fill=Tk.X, expand=1, padx=10, pady=20)

        label = Tk.Label(t_frame, text="Number of Worker Processes: ")
        label.grid(row=0, column=0)
        label = Tk.Label(t_frame, text="%d" % self.num_workers, bg='white')
        label.grid(row=0, column=1)
        if self.use_gpu:
            label = Tk.Label(t_frame, text="Using GPU", bg='white')
            label.grid(row=0, column=2, padx=20)

        s_frame = ScrolledFrame(body_frame)
        s_frame.pack(padx=10, pady=5)
        i_frame = s_frame.interior

        items = ['Job', 'Analysis Name', 'File Name', 'PID', 'State', 'Progress (max=%d)' % MAXNUM_STEPS, 'Step', 'Chi2', 'Rg', 'Support Volume', 'Submitted', 'Started', 'Finished', 'Duration', 'Cancel']

        for j, item in enumerate(items):
            label = Tk.Label(i_frame, text=item)
            label.grid(row=0, column=j)

        widths = [30, 80, 80, 40, 60, PROGRES_BAR_WIDTH, 50, 50, 50, 50, 60, 60, 60, 40, 40]
        self.total_width = np.sum(widths)
        for j, w in enumerate(widths):
            space = Tk.Frame(i_frame, width=w)
            space.grid(row=1, column=j)

        self.widget_table = []
        self.i_frame = i_frame

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack(fill=Tk.X)

        w = Tk.Button(box, text="Cancel All", width=10, command=self.cancel_all_jobs, fg='red', bg='khaki')
        w.pack(side=Tk.RIGHT, padx=40, pady=5)
        self.calcel_all_btn = w

        f = Tk.Frame(box)
        f.pack(side=Tk.RIGHT, fill=Tk.X, expand=1)

        w = Tk.Button(f, text="Close and Keep", width=14, command=self.ok, default=Tk.ACTIVE)
        w.pack()

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def update_geometry(self, num_jobs):
        self.update()
        btn_w, btn_h, btn_x, btn_y = get_widget_geometry(self.calcel_all_btn)
        # w, h, x, y = get_widget_geometry(self)

        print('update_geometry', btn_w, btn_h, btn_x, btn_y)
        # print('update_geometry', w, h, x, y)
        w_ = int(btn_w * 14.5)
        h_ = btn_h*(num_jobs+7)
        self.geometry('%dx%d' % (w_, h_))
        self.update()

    def update_table(self):
        from .DenssManager import get_list
        job_list = get_list()
        if False:
            print(strftime("%H:%M:%S"),'update_table')
            for rec in job_list:
                print(rec)

        if self.basic_geometry is None:
            # w, h, x, y
            self.basic_geometry = get_widget_geometry(self.calcel_all_btn)

        i_frame = self.i_frame
        bg_color = 'white'

        need_refresh = False
        for i, row_rec in enumerate(job_list):
            job_id, job_rec = row_rec
            analysis, file = job_rec[0]
            pid = job_rec[1]
            state = job_rec[2]
            step = job_rec[3]
            chi2 = job_rec[4]
            rg = job_rec[5]
            vol = job_rec[6]
            submitted = job_rec[7]
            started = job_rec[8]
            finished = job_rec[9]
            duration = job_rec[10]

            if i < len(self.widget_table):
                widget_rec = self.widget_table[i]

                if state > -3:
                    need_refresh = True

                update_duration = False
                if pid is None:
                    if state == 0:
                        pass
                    elif state <= -3:
                        finished_label = widget_rec[12]
                        text = finished_label.cget('text')
                        if text == '':
                            pid_label = widget_rec[3]
                            pid_label.config(text='')
                            finished_label.config(text=format_time(finished))
                            state_ = get_state_text(state)
                            color = get_state_color(state)
                            state_label = widget_rec[4]
                            state_label.config(text=state_, fg=color)
                            cancel_button = widget_rec[14]
                            cancel_button.config(state=Tk.DISABLED)
                            update_duration = True
                else:
                    pid_label = widget_rec[3]
                    text = pid_label.cget('text')
                    if text == "":
                        pid_label.config(text=str(pid))

                    if state < 0 or step < 100:
                        state_ = get_state_text(state)
                        color = get_state_color(state)
                        state_label = widget_rec[4]
                        state_label.config(text=state_, fg=color)
                        for k, time_ in enumerate([submitted, started]):
                            text = strftime("%H:%M:%S", localtime(time_))
                            time_label = widget_rec[k+10]
                            time_label.config(text=text)
                        cancel_button = widget_rec[14]
                        cancel_button.config(state=Tk.DISABLED if state <= -3 else Tk.NORMAL)

                    update_duration = True

                if state < 0:
                    if state == -1:
                        update_duration = True
                else:
                    step_label = widget_rec[6]
                    step_label.config(text=VALUE_FORMATS[0] % step)
                    chi2_label = widget_rec[7]
                    fg_color = get_ch2_fg_color(chi2)
                    chi2_label.config(text=VALUE_FORMATS[1] % chi2, fg=fg_color)
                    rg_label = widget_rec[8]
                    rg_label.config(text=VALUE_FORMATS[2] % rg)
                    vol_label = widget_rec[9]
                    vol_label.config(text=VALUE_FORMATS[3] % vol)

                if update_duration:
                    duration_label = widget_rec[13]
                    duration_label.config(text=format_duration(duration))

                pbar = widget_rec[5]
                if step >= 0:
                    pbar["value"] = step
            else:
                need_refresh = True
                widget_rec = []
                row = i+1

                # [0] Job
                col = 0
                label = Tk.Label(i_frame, text="%03d" % job_id, bg=bg_color)
                label.grid(row=row, column=col)
                widget_rec.append(label)

                # [1] Analysis Name
                col +=1
                label = Tk.Label(i_frame, text=analysis, bg=bg_color)
                label.grid(row=row, column=col, padx=5)
                widget_rec.append(label)

                # [2] File Name
                col +=1
                label = Tk.Label(i_frame, text=file, bg=bg_color)
                label.grid(row=row, column=col)
                widget_rec.append(label)

                # [3] PID
                col +=1
                label = Tk.Label(i_frame, text="" if pid is None else str(pid), bg=bg_color)
                label.grid(row=row, column=col, padx=5)
                widget_rec.append(label)

                # [4] State
                col +=1
                color = get_state_color(state)
                label = Tk.Label(i_frame, text=get_state_text(state), bg=bg_color, fg=color)
                label.grid(row=row, column=col)
                widget_rec.append(label)

                # [5] Progress
                col +=1
                length = self.basic_geometry[0] * PROGRES_BAR_SCALE
                pbar = ttk.Progressbar(i_frame, orient ="horizontal", length=length, mode="determinate")
                pbar.grid(row=row, column=col, padx=10)
                pbar["maximum"] = MAXNUM_STEPS
                step_ = MAXNUM_STEPS if step == -3 else step
                pbar["value"] = max(0, step_)
                widget_rec.append(pbar)

                # [6] Step, [7] Chi2, [8] Rg, [9] Support Volume
                for k, value in enumerate([step, chi2, rg, vol]):
                    col +=1
                    text = "" if value is None or value < 0 else VALUE_FORMATS[k] % value
                    if k == 1:  # i.e., Ch2
                        fg_color = get_ch2_fg_color(value)
                    else:
                        fg_color = None
                    label = Tk.Label(i_frame, text=text, bg=bg_color, fg=fg_color)
                    label.grid(row=row, column=col, padx=5)
                    widget_rec.append(label)

                # [10] Submitted
                col +=1
                label = Tk.Label(i_frame, text=format_time(submitted), bg=bg_color)
                label.grid(row=row, column=col)
                widget_rec.append(label)

                # [11] Started
                col +=1
                label = Tk.Label(i_frame, text=format_time(started), bg=bg_color)
                label.grid(row=row, column=col, padx=5)
                widget_rec.append(label)

                # [12] Finished
                col +=1
                label = Tk.Label(i_frame, text=format_time(finished), bg=bg_color)
                label.grid(row=row, column=col)
                widget_rec.append(label)

                # [13] Duration
                col +=1
                label = Tk.Label(i_frame, text=format_duration(duration), bg=bg_color)
                label.grid(row=row, column=col)
                widget_rec.append(label)

                # [14] Cancel
                col +=1
                state_ = Tk.DISABLED if pid is None and state <= -3 else Tk.NORMAL
                button = Tk.Button(i_frame, text="Cancel", fg='red', bg='khaki', state=state_, command=lambda j_=job_id :self.cancel_job(j_))
                button.grid(row=row, column=col)
                widget_rec.append(button)

                self.widget_table.append(widget_rec)

        if need_refresh:
            if not self.repeat_refresh:
                self.repeat_refresh = True
                self.after(REFRESH_INTERVAL, self.refresh)
        else:
            self.repeat_refresh = False

        self.calcel_all_btn.config(state=Tk.NORMAL if self.repeat_refresh else Tk.DISABLED)

    def submit_jobs(self, jobs):
        from .DenssManager import submit
        for job_info in jobs:
            submit(job_info)

    def refresh(self):
        self.update_table()
        if self.repeat_refresh:
            self.after(REFRESH_INTERVAL, self.refresh)

    def apply(self):
        # print('apply')
        self.repeat_refresh = False
        self.update()

    def cancel_job_impl(self, job_id):
        from .DenssManager import get_table
        table = get_table()
        last_rec = table[job_id]
        last_rec[2] = -5    # canceled
        last_rec[11] = 1
        table[job_id] = last_rec

    def cancel_job(self, job_id):
        print('cancel_job', job_id)
        import molass_legacy.KekLib.CustomMessageBox as MessageBox
        yn = MessageBox.askyesno(
                'Cancel Job Confirmation',
                'Do you really want to cancel job %03d ?' % job_id,
                parent=self,
                )
        if not yn:
            return

        self.cancel_job_impl(job_id)

    def cancel_all_jobs(self):
        import molass_legacy.KekLib.CustomMessageBox as MessageBox
        yn = MessageBox.askyesno(
                'Cancel All Confirmation',
                'Do you really want to cancel all jobs?',
                parent=self,
                )
        if not yn:
            return

        from .DenssManager import get_list
        job_list = get_list()
        for i, row_rec in enumerate(job_list):
            job_id, job_rec = row_rec
            state = job_rec[2]
            if state >= -2:
                self.cancel_job_impl(job_id)



# coding: utf-8
"""
    MixedElutionModel.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from sklearn.mixture import GaussianMixture
from sklearn.cluster import KMeans
import molass_legacy.KekLib.DebugPlot as plt

USE_DPHEM_REDUCTIONE = False

def guess_n_components(x, y, p=0.3):
    if p is None:
        x_ = x
        y_ = y
    else:
        kth = int(len(y)*p)
        pp = np.argpartition(y, kth)
        pp_ = pp[kth:]
        x_ = x[pp_]
        y_ = y[pp_]

    X_ = np.expand_dims(y_,1)
    scores = []
    kmax = 8

    kx = np.arange(2, kmax+1)
    for k in kx:
      gmm = GaussianMixture(n_components=k, max_iter=40).fit(X=X_)
      scores.append((gmm.aic(X_), gmm.bic(X_)))

    scores = np.array(scores)

    if True:

        plt.push()

        fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(18,5))
        ax1, ax2, ax3 = axes

        ax1.set_title("%.2g percentile discarded" % (p*100))
        ax2.set_title("Information Criterions")
        ax1.plot(x, y, 'o')
        ax1.plot(x_, y_, 'o')
        ax2.plot(kx, scores[:,0], '-o', label='AIC')
        ax2.plot(kx, scores[:,1], '-o', label='BIC')
        ax2.legend()
        fig.tight_layout()
        plt.show()

        plt.pop()

    k = np.argmin(scores[:,1])
    return kx[k]

class MixedElutionModel(GaussianMixture):
    def __init__(self, **kwargs):
        GaussianMixture.__init__(self, **kwargs)

def spike_demo(in_folder, lpm_correct=False, eslice=None):
    from Prob.GaussianMixture import gm_curve, hist_to_source, debug_plot_gmm
    from Prob.gmm_dphem_learn import gmm_dphem_learn

    if in_folder is None:
        x, y = gm_curve()
    else:
        from RawData import RawXray
        from molass_legacy.Saxs.RankAnalysis import RankAnalysis
        rx = RawXray(in_folder)
        i = rx.get_row_index(0.02)
        x = np.arange(rx.data.shape[1])
        y = rx.data[i,:]
        if lpm_correct:
            from LPM import get_corrected
            y = get_corrected(y)
        if eslice:
            y = y[eslice]
            x = np.arange(len(y))

    sy = hist_to_source(x, y)

    k = guess_n_components(x, y)
    print('k=', k)

    temp_gmm = MixedElutionModel(n_components=k)
    temp_gmm.fit(X=np.expand_dims(sy,1))

    if USE_DPHEM_REDUCTIONE:
        tagk = 2
        opt = {'pimode': 'preserve', 
                'initmode': 'random'}
        result_gmm, post, info, diffel = gmm_dphem_learn(temp_gmm,tagk,opt)
    else:
        result_gmm = temp_gmm
        temp_gmm = None

    debug_plot_gmm(x, y, sy, result_gmm, gmm2=temp_gmm)



# coding: utf-8
"""
    TrimmingInspection.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import logging
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from DataUtils import get_in_folder

MARGIN_WIDTH = 0.01     # angstrom⁻¹

class ElutionTrimmingDialog(Dialog):
    def __init__(self, parent, dialog, data, ecurve, trimming_info):
        self.ecurve = ecurve
        self.trimming_info = trimming_info
        Dialog.__init__(self, parent, "Elution Trimming Inspector", visible=False, location="lower right")

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()

        fig = plt.figure(figsize=(12,6))
        self.fig = fig

        ax1 = fig.add_subplot(111)

        self.draw_elution(ax1)

        fig.tight_layout()
        # fig.subplots_adjust(top=0.85, bottom=0.1)

        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)

        tframe = Tk.Frame(body_frame)
        tframe.pack(padx=0, fill=Tk.X)      # padx=0: suppress changes from toolbar cursor display
        tframe_left = Tk.Frame(tframe)
        tframe_left.pack(side=Tk.LEFT, fill=Tk.X)
        tframe_right = Tk.Frame(tframe)
        tframe_right.pack(side=Tk.RIGHT)

        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe_left)
        self.toolbar.update()

    def cancel(self):
        plt.close(self.fig)
        Dialog.cancel(self)

    def draw_elution(self, ax):
        from molass_legacy.Elution.CurveUtils import simple_plot

        in_folder = get_in_folder()
        ax.set_title("Elution Trimming Inspection on %s" % in_folder, fontsize=20)
        ecurve = self.ecurve

        simple_plot(ax, ecurve, color="orange", legend=False)

        trinfo = self.trimming_info
        extra = trinfo.extra

        if extra is not None:

            ymin, ymax = ax.get_ylim()
            ax.set_ylim(ymin, ymax)

            emg_peaks = [extra.info_L[0]]
            if extra.num_peaks > 1:
                emg_peaks.append(extra.info_R[0])

            ex = ecurve.x
            for peak in emg_peaks:
                y = peak.get_model_y(ex)
                ax.plot(ex, y, ":", color="blue", label="original elution model", lw=2)

            sinfo = extra.sigma_info
            for k, px in enumerate([sinfo.jmin, sinfo.jmax]):
                label = "original range" if k == 0 else None
                ax.plot([px, px], [ymin, ymax], color="blue", label=label)

            recon_info = self.trimming_info.extra.recon_info
            if recon_info is not None:
                rem_y = recon_info.prm_model(ex)
                ax.plot(ex, rem_y, ":", color="red", label="revised elution model", lw=2)

                for k, j in enumerate([recon_info.jmin, recon_info.jmax]):
                    label = "revised range" if k == 0 else None
                    ax.plot([j, j], [ymin, ymax], color="red", label=label)

            for k, px in enumerate([trinfo.start, trinfo.stop]):
                label = "final range" if k == 0 else None
                ax.plot([px, px], [ymin, ymax], ":", color="yellow", label=label, lw=2)

        ax.legend()

    def save_the_figure(self, file):
        self.fig.savefig(file)



# coding: utf-8
"""
    Uv2dLpm.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.Baseline.ScatteringBaseline import ScatteringBaseline
import molass_legacy.KekLib.DebugPlot as plt

class Uv2dLpm:
    def __init__(self, a_vector, iy):
        self.iy = iy
        y = a_vector[iy]
        sbl = ScatteringBaseline(y)
        A, B = sbl.solve()

        if False:
            x = np.arange(iy.start, iy.stop)
            fig = plt.figure()
            ax = fig.gca()
            ax.plot(a_vector)
            ax.plot(x, y)
            ax.plot(x, A*x + B)
            fig.tight_layout()
            plt.show()

        self.params = (A, B)

    def get_params(self):
        # return params interpreted as in 3d
        A, B = self.params
        b = B - self.iy.start*A
        return 0, A, b



"""
    ATSAS.AutoRg.py

    extracted from SerialAtsasTools.py

    Copyright (c) 2016-2022, SAXS Team, KEK-PF
"""
import os
import glob

USED_IN_MOLASS = True
if USED_IN_MOLASS:
    from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
else:
    from molass_legacy.AutorgKek.Settings import get_setting, set_setting

autorg_exe_array = []

def get_dirs_in_later_order(pattern):
    paths = glob.glob(pattern)
    paths.sort(reverse=True)
    return paths

def get_autorg_exe_paths():
    dir_patterns = [r'C:\Program Files\ATSAS*', r'C:\Program Files (x86)\ATSAS*', r'C:\atsas*']
    atsas_dirs = []

    try:
        current_dir = os.environ["ATSAS"]
        if os.path.exists(current_dir):
            current_dir = current_dir.replace("/", "\\")
            atsas_dirs.append(current_dir)
    except:
        pass

    for pattern in dir_patterns:
        atsas_dirs += get_dirs_in_later_order(pattern)

    autorg_exe_array = []
    for dir_ in atsas_dirs:
        for sub_dir in [ r'\bin', '' ]:
            exe_path = dir_ + sub_dir + r'\autorg.exe'
            if os.path.exists(exe_path) and exe_path not in autorg_exe_array:
                autorg_exe_array.append( exe_path )
                break
    return autorg_exe_array

def set_exe_array(paths=None):
    if paths is None:
        temp_paths = get_setting( 'atsas_exe_paths' )
        if len(temp_paths) == 0:
            temp_paths = get_autorg_exe_paths()
    else:
        temp_paths = paths

    autorg_exe_array.clear()
    for path in temp_paths:
        autorg_exe_array.append(path)

    set_setting( 'atsas_exe_paths', autorg_exe_array )

set_exe_array()

def reset_exe_array():
    global autorg_exe_array
    autorg_exe_array = get_autorg_exe_paths()



"""
    CfsEvalPeaks.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from scipy.interpolate import UnivariateSpline

CFSE_WEIGHT = 5

def get_cfs_evaluated_peaks(x, y, num_peaks, using_hybrid=False, using_cfs=True, debug=False):
    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
    from molass_legacy.QuickAnalysis.ModeledPeaks import recognize_peaks
    from molass_legacy.Models.Characteristic import CfSpace
    from molass_legacy.Models.ElutionCurveModels import egha, emga
    from molass_legacy.Models.Egh2Emg import to_emga_params
    from SecTheory.Edm import edm_func
    from SecTheory.EdmSpike import guess_edm_initial_params

    if using_hybrid or using_cfs:
        cfs = CfSpace()
        cft0 = cfs.compute_cf(x, y)

    try_num_peaks = num_peaks + 1
    egha_params_list = recognize_peaks(x, y, exact_num_peaks=try_num_peaks, affine=True)

    if debug:
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
        ax1.plot(x, y)

    shape = (try_num_peaks, 5)

    def objective_egha_cfs(params, debug=False):
        cy_list = []
        for p in params.reshape(shape):
            cy = egha(x, *p)
            cy_list.append(cy)
        ty = np.sum(cy_list, axis=0)

        if debug:
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.plot(x, y)
                for cy in cy_list:
                    ax.plot(x, cy, ":")
                ax.plot(x, ty, ":", color="red")
                fig.tight_layout()
                plt.show()

        if using_hybrid:
            cft1 = cfs.compute_cf(x, ty)
            return np.log(np.sum((ty - y)**2)) + CFSE_WEIGHT*np.log(np.sum((cft1 - cft0)**2))
        elif using_cfs:
            cft1 = cfs.compute_cf(x, ty)
            return np.sum((cft1 - cft0)**2)
        else:
            return np.sum((ty - y)**2)

    params = np.array(egha_params_list).flatten()
    # objective_egha_cfs(params, debug=True)

    ret_egha = minimize(objective_egha_cfs, params)
    # objective_egha_cfs(ret_egha.x, debug=True)

    def objective_emga_cfs(params, debug=False):
        ty = np.zeros(len(x))
        for p in params.reshape(shape):
            cy = emga(x, *p)
            ty += cy

        if debug:
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.plot(x, y)
                ax.plot(x, ty)
                fig.tight_layout()
                plt.show()

        if using_hybrid:
            cft1 = cfs.compute_cf(x, ty)
            return np.log(np.sum((ty - y)**2)) + CFSE_WEIGHT*np.log(np.sum((cft1 - cft0)**2))
        elif using_cfs:
            cft1 = cfs.compute_cf(x, ty)
            return np.sum((cft1 - cft0)**2)
        else:
            return np.sum((ty - y)**2)

    emga_params_list = []
    for params in ret_egha.x.reshape(shape):
        emga_params_list.append(to_emga_params(params))

    params = np.array(emga_params_list).flatten()
    # objective_emga_cfs(params, debug=True)

    ret_emga = minimize(objective_emga_cfs, params)
    # objective_emga_cfs(ret_emga.x, debug=True)

    edm_shape = (try_num_peaks, 6)

    def objective_edm_cfs(params, debug=False):
        ty = np.zeros(len(x))
        for p in params.reshape(edm_shape):
            cy = edm_func(x, *p)
            ty += cy

        if debug:
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.plot(x, y)
                ax.plot(x, ty)
                fig.tight_layout()
                plt.show()

        if using_hybrid:
            cft1 = cfs.compute_cf(x, ty)
            return np.log(np.sum((ty - y)**2)) + CFSE_WEIGHT*np.log(np.sum((cft1 - cft0)**2))
        elif using_cfs:
            cft1 = cfs.compute_cf(x, ty)
            return np.sum((cft1 - cft0)**2)
        else:
            return np.sum((ty - y)**2)

    init_edms = guess_edm_initial_params(x, y)
    u = 0.5
    edm_params_list = []
    for p in init_edms:
        edm_params_list.append((u, *p))

    params = np.array(edm_params_list).flatten()
    # objective_edm_cfs(params, debug=True)

    ret_edm = minimize(objective_edm_cfs, params)
    # objective_edm_cfs(ret_edm.x, debug=True)

    return ret_egha.x.reshape(shape), ret_emga.x.reshape(shape), ret_edm.x.reshape(edm_shape)



"""
main module
"""
import numpy as np
from . import moments as mm
from .model_gm import ModelGM

    # input: samples
    # output: estimated model
    # provide some optional parameters


    # known variance
    ## given weighting matrix
    ## no weighting matrix
    ### 1. use identity weighting matrix
    ### 2. consistent estimation of optimal weighting matrix
    ### 3. re-estimate


    # unknown variance
class DMM():
    """
    class for denoised method of moments
    """
    def __init__(self, k, interval=None, sigma=None):
        """
        Args:
        sigma: float, default None
        standard deviation. If sigma == None, will estimate sigma.

        interval =[a,b]: floats, default [-10, 10]
        represents the interval of means [a,b]

        num_components: int, required
        number of postulated components
        """
        if interval is None:
            interval = [-10, 10]

        self.sigma = sigma
        self.k = k
        self.interval = interval

        if self.sigma is None:
            self.num_mom = 2*self.k
        else:
            self.num_mom = 2*self.k-1
            self.transform = mm.hermite_transform_matrix(2*self.k-1, self.sigma)

        # for online estimation
        # if sigma is None, this stores list of (sample size, raw moments)
        # if sigma is known, this store list of (sample size, Hermite moments, correlation)
        if self.sigma is None:
            self.online = [0, np.zeros((self.num_mom, 1))]
        else:
            self.online = [0, np.zeros((self.num_mom, 1)), np.zeros((self.num_mom, self.num_mom))]


    def estimate(self, samples):
        """
        estimate a model from given samples
        use two-step estimate:
        1.(a) prelimnary estimation with identity weight matrix
          (b) estimation of optimal weight matrix (require all samples)
        2.    reestimate parameters using esimated weight matrix

        Args:
        samples: array of float, required
        samples collected

        Returns:
        an estimated ModelGM
        """
        samples = np.asarray(samples)
        m_raw = mm.empirical_moment(samples, self.num_mom)

        if self.sigma is None:
            m_raw = np.mean(m_raw, axis=1)
            m_esti, var_esti = mm.deconvolve_unknown_variance(m_raw)
            dmom_rv = mm.quadmom(m_esti[:2*self.k-1])
            return ModelGM(w=dmom_rv.weights, x=dmom_rv.atoms, std=np.sqrt(var_esti))
        else:
            m_hermite = np.dot(self.transform[0], m_raw)+self.transform[1]
            m_decon = np.mean(m_hermite, axis=1)
            dmom_rv = self.estimate_from_moments(m_decon) # preliminary estimate
            wmat = estimate_weight_matrix(m_hermite, dmom_rv)
            dmom_rv = self.estimate_from_moments(m_decon, wmat) # second step estimate
            # print(np.linalg.inv(wmat))
            return ModelGM(w=dmom_rv.weights, x=dmom_rv.atoms, std=self.sigma)

    def estimate_online(self, samples_new):
        """
        update the estimate a model from more samples
        only store a few moments and correlations

        Args:
        samples_new: array of floats
        new samples

        Returns:
        an estimated ModelGM
        """
        samples_new = np.asarray(samples_new)
        m_new = mm.empirical_moment(samples_new, self.num_mom) # moments, shape (L,n)
        n_new = len(samples_new)
        n_total = self.online[0]+n_new

        if self.sigma:
            m_new = np.dot(self.transform[0], m_new)+self.transform[1]
            cor_new = np.dot(m_new, m_new.T)/n_new
            self.online[2] = self.online[2]*(self.online[0]/n_total)+cor_new*(n_new/n_total)

        mom_new = np.mean(m_new, axis=1)[:, np.newaxis] # empirical moments, shape (L,1)
        self.online[1] = self.online[1]*(self.online[0]/n_total)+mom_new*(n_new/n_total)
        self.online[0] = n_total

        if self.sigma is None:
            m_esti, var_esti = mm.deconvolve_unknown_variance(self.online[1])
            dmom_rv = mm.quadmom(m_esti[:2*self.k-1])
            return ModelGM(w=dmom_rv.weights, x=dmom_rv.atoms, std=np.sqrt(var_esti))
        else:
            wmat = np.linalg.inv(self.online[2]-np.dot(self.online[1], self.online[1].T))
            dmom_rv = self.estimate_from_moments(self.online[1].reshape(self.num_mom), wmat)
            # print(np.linalg.inv(wmat))
            return ModelGM(w=dmom_rv.weights, x=dmom_rv.atoms, std=self.sigma)

    def estimate_with_wmat(self, samples, wmat=None):
        """
        estimate a model from given samples using given weight matrix
        model: X=U+sigma*Z
        sigma must be given

        Args:
        samples: array of float, required
        samples collected

        wmat: array of shape (k,k)
        weight matrix, default identity

        Returns:
        latent distribtuion
        """
        assert self.sigma

        m_latent = self.estimate_latent_moments(samples)
        dmom_rv = self.estimate_from_moments(m_latent, wmat)
        return ModelGM(w=dmom_rv.weights, x=dmom_rv.atoms, std=self.sigma)

    def estimate_select(self, samples, threhold=1):
        """
        estimate with selected number of components
        """
        k_cur = min(self.select_num_comp(samples, threhold), self.k)
        dmm_cur = DMM(k=k_cur, interval=self.interval, sigma=self.sigma)
        return dmm_cur.estimate(samples)

    def estimate_latent_moments(self, samples):
        """
        estimate moments of latent distribution (deconvolution)
        model: X=U+sigma*Z
        sigma must be given

        Args:
        samples: array of length n

        Return:
        array of length 2k-1
        estimated moments of U from degree 1 to 2k-1
        """
        assert self.sigma

        samples = np.asarray(samples)
        m_raw = mm.empirical_moment(samples, self.num_mom)
        m_raw = np.mean(m_raw, axis=1).reshape((self.num_mom, 1))
        return ((np.dot(self.transform[0], m_raw)+self.transform[1])).reshape(self.num_mom)

    def estimate_from_moments(self, moments, wmat=None):
        """
        estimate a discrete random variable from moments estimate

        Args:
        moments: array of length 2k-1
        estimated moments of U of degree 1 to 2k-1

        wmat: matrix of shape (k, k)
        weight matrix for moment projection, default identity matrix

        Returns:
        an estimated latent distribtuion on at most k points
        """
        m_proj = mm.projection(moments, self.interval, wmat)
        dmom_rv = mm.quadmom(m_proj, dettol=0)
        return dmom_rv

    def sample_moment_cov(self, samples):
        """
        return the sample covariance matrix of moments estimates
        """
        samples = np.asarray(samples)
        mom = mm.empirical_moment(samples, self.num_mom) # moments, shape (L,n)
        mean = np.mean(mom, axis=1)
        num = len(samples)
        cor = np.dot(mom, mom.T)/num - np.outer(mean, mean)
        return cor

    def select_num_comp(self, samples, threhold=1):
        """
        select the number of components
        according to sample variance of moments estimate
        """
        samples = np.asarray(samples)
        num = len(samples)

        m_raw = np.ones(len(samples))
        deg_cur = 0
        while True:
            deg_cur += 1
            m_raw *= samples
            var = np.mean(m_raw**2)-np.mean(m_raw)**2
            if var > threhold*num:
                break

        # moments of degree 1 to (deg_cur-1) is accurate
        if self.sigma:
            return int(np.floor(deg_cur/2))
        else:
            return int(np.floor((deg_cur-1)/2))

def estimate_weight_matrix(m_estimate, model):
    """
    estimate weight matrix: inverse of the estimated covariance matrix
    ref: [Bruce E. Hansen] Econometrics. Chapter 11.

    Args:
    m_estimate: matrix of size (k,n)
    power of n samples from degree of 1 to k

    model: discrete_rv

    Return:
    consistent estimation for the optimal weight matrix
    """
    num_moments, num_samples = m_estimate.shape
    mom_model = model.moment(num_moments).reshape((num_moments, 1))
    m_cond = m_estimate - mom_model
    m_cond_avg = np.mean(m_cond, axis=1).reshape((num_moments, 1))
    m_cond_centd = m_cond - m_cond_avg
    return np.linalg.inv(np.dot(m_cond_centd, m_cond_centd.T)/num_samples)



# coding: utf-8
"""

    ファイル名：    AppVersion.py

    処理内容：      アプリケーションのバージョン情報


"""
from __future__ import division, print_function, unicode_literals

import platform

def get_com_version_string():
    return 'autorg_kek 0.7.1 (2019-04-18 python %s %s)' % ( platform.python_version(), platform.architecture()[0] )

def autoguinier_version_for_publication():
    import re
    version = get_com_version_string().replace("autorg_kek", "AutoGuinier")
    return re.sub(r"\s+\(.+", "", version)



"""
    SimTools.ErrorModel.py - revised version of Error.ErrorModel.py 

    based on 2017, Steffen M. Sedlak, et al.
    Quantitative evaluation of statistical errors in smallangle X-ray scattering measurements

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
from scipy.optimize import minimize, curve_fit
from scipy.interpolate import UnivariateSpline
from matplotlib.gridspec import GridSpec
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Peaks.ElutionModels import egh

def error_model(q, Iq, k, c, Iarb):
    return 1/(k*q)*(Iq + 2*c*Iarb/(1 - c))      # Eq.(10) in the paper

K_SCALE = 1e6

class ErrorModel:
    def __init__(self, qv, D, E, spline=None):
        self.E = E

        i = bisect_right(qv, 0.02)
        cy = np.average(D[i-5:i+6,:], axis=0)
        j = np.argmax(cy)
        sy = np.average(D[:,j-5:j+6], axis=1)
        se = np.average(E[:,j-5:j+6], axis=1)

        slice_ = slice(10, -10)
        # slice_ = slice(None, None)
        qv_ = qv[slice_]
        sy_ = sy[slice_]
        se_ = se[slice_]
        se2 = se_**2

        def obj_func(p):
            k, c = p
            y_ = error_model(qv_, sy_, k*K_SCALE, c, sy[i])
            return np.sum(((y_ - se2)/y_)**2)

        ret = minimize(obj_func, (1, 0.5), method="Nelder-Mead")
        k_, c_ = ret.x
        print("k_, c_=", k_, c_)
        print("ret.success=", ret.success)
        print("ret.nit=", ret.nit)

        sem =  error_model(qv_, sy_, k_*K_SCALE, c_, sy[i])

        if False:
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_yscale("log")
                ax.plot(qv_, se_)
                ax.plot(qv_, np.sqrt(sem))
                fig.tight_layout()
                plt.show()

        if spline is None:
            spline = UnivariateSpline(qv, sy, s=0, ext=3)

        self.spline = spline
        self.Iarb = sy[i]
        self.params = ret.x

    def get_error_curve(self, qv, scale=1):
        k, c = self.params
        return np.sqrt(error_model(qv, scale*self.spline(qv), k*K_SCALE, c, scale*self.Iarb))

def demo(in_folder, v2_copy):
    from molass_legacy.DENSS.DenssUtils import fit_data_bc
    from DataUtils import get_in_folder
    from MatrixData import simple_plot_3d
    from molass_legacy.KekLib.BasicUtils import Struct
    from SvdDenoise import get_denoised_data

    D, E, qv, ecurve = v2_copy.get_xr_data_separate_ly()

    i = bisect_right(qv, 0.02)

    cx = np.arange(D.shape[1])
    cy = np.average(D[i-5:i+6,:], axis=0)

    p_init = ecurve.get_emg_peaks()[ecurve.primary_peak_no].get_params()
    popt, pcov = curve_fit(egh, cx, cy, p_init)
    cym = egh(cx, *popt)

    C = np.array([cym, cym**2])
    D_ = get_denoised_data(D, rank=2)
    P = D_ @ np.linalg.pinv(C)

    k = np.argmax(cym)

    aq, bq = P.T
    aq_ = aq*cym[k]

    j = np.argmax(cy)
    sy = np.average(D[:,j-5:j+6], axis=1)
    ey = np.average(E[:,j-5:j+6], axis=1)

    fq, fy, fe, _ = fit_data_bc(qv, sy, ey)
    spline = UnivariateSpline(fq, fy, s=0, ext=3)

    fq_, fy_, fe_, _ = fit_data_bc(qv, aq_, ey)

    em = ErrorModel(qv, D, E, spline=spline)
    eym = em.get_error_curve(qv)

    in_folder = get_in_folder(in_folder)
    super_title = "Error Model for %s based on 2017, Steffen M. Sedlak, et al." % in_folder

    def demo_plot():

        with plt.Dp():
            # fig, axes = plt.subplots(nrows=2, ncols=3, figsize=(15,10))
            # ax1, ax2, ax3 = axes[0,:]
            fig = plt.figure(figsize=(15,10))
            ax1 = fig.add_subplot(231)
            ax2 = fig.add_subplot(232)
            ax3 = fig.add_subplot(233)
            ax4 = fig.add_subplot(234, projection="3d")
            ax5 = fig.add_subplot(235, projection="3d")
            ax6 = fig.add_subplot(236, projection="3d")

            fig.suptitle(super_title, fontsize=20)
            ax1.set_title("Elution at Q=%g" % 0.02, fontsize=16)
            ax2.set_title("Scattering Intensity at j=%d" % j, fontsize=16)
            ax3.set_title("Scattering Error at j=%d" % j, fontsize=16)

            for ax in [ax2, ax3]:
                ax.set_yscale("log")
            ax1.plot(cx, cy, label="data")
            ax1.plot(cx, cym, label="egh")
            ax1.legend()

            ax2.plot(qv, sy, label="data")
            ax2.plot(fq, fy, ":", lw=2, label="IFT (P. B. Moore 1980) from data")
            ax2.plot(qv, aq_, label="A(q)")
            ax2.plot(fq, fy_, ":", lw=2, label="IFT (P. B. Moore 1980) from A(q)")
            ax2.legend()

            ax3.plot(qv, ey, label="data")
            ax3.plot(qv, eym, label="model")
            ax3.legend()

            ax4.set_zscale("log")
            simple_plot_3d(ax4, E, x=qv)

            fig.tight_layout()
            plt.show()

    demo_plot()

    from importlib import reload

    import Trials.BoundedLRF.IterativeLrfSolver
    reload(Trials.BoundedLRF.IterativeLrfSolver)

    from Trials.BoundedLRF.IterativeLrfSolver import IterativeLrfSolver

    f = 130
    t = 200
    range_ = slice(f, t+1)

    solver = IterativeLrfSolver(qv, D[:,range_], E[:,range_])
    P, C, Rg, R, L, bq_bounds, coerced_bq, changes = solver.solve(maxiter=100000, L=0.4)

    old_aq = aq_

    aq, bq = P.T
    aq_ = aq*cym[k]
    fq_, fy_, fe_, _ = fit_data_bc(qv, aq_, ey)

    with plt.Dp():
        fig, ax = plt.subplots()
        ax.set_yscale("log")
        ax.plot(qv, sy, label="data")
        ax.plot(qv, old_aq, label="old A(q)")
        ax.plot(qv, aq_, label="new A(q)")
        ax.legend()
        fig.tight_layout()
        plt.show()

    super_title = "Result of Bounded LRF for %s" % in_folder
    demo_plot()



"""
    Simulative/LognormalPsd.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from matplotlib.gridspec import GridSpec
from matplotlib.widgets import Slider, Button

def lognormalpore_model_interactive_impl(x, y, params, rgs,
                                         title=None, use_ty_as_data=False, plot_ty=None,
                                         plot_mnp=True,
                                         window_title=None, parent=None):
    import molass_legacy.KekLib.DebugPlot as plt
    from molass_legacy.KekLib.MatplotlibUtils import get_labeltext_from_line
    from molass_legacy.Models.Stochastic.LognormalPoreFunc import lognormal_pore_pdf, distr_func, compute_trvec
    from molass_legacy.Models.Stochastic.LognormalUtils import compute_mode, compute_stdev, compute_mu_sigma
    from SecTheory.BasicModels import robust_single_pore_pdf


    if plot_ty is None:
        plot_ty = not use_ty_as_data

    N, T, t0, me, mp, mu, sigma = params[0:7]
    scales = params[7:]

    rgv = np.flip(np.linspace(10, 200, 100))
    trv = compute_trvec(N, T, t0, me, mp, mu, sigma, rgv)
    # plot_rv_list()

    with plt.Dp(parent=parent, window_title=window_title, button_spec=["OK", "Cancel"]):
        fig = plt.figure(figsize=(20,4.5))
        gs = GridSpec(1,4)
        ax1 = fig.add_subplot(gs[0,0])
        ax2 = fig.add_subplot(gs[0,1:3])
        ax3 = fig.add_subplot(gs[0,3])
    
        if title is None:
            title = "Simulative Use Demo with Lognormalpore vs. Monopore Comparison"
        fig.suptitle(title, fontsize=20)
        ax1.set_title("Experiment Data", fontsize=16)
        ax2.set_title("Simulated Data", fontsize=16)
        ax3.set_title("Pore Size Distribution", fontsize=16)

        def compute_lnp_curves(x, N, T, me, mp, mu, sigma, t0):
            cy_list = []
            for rg, scale in zip(rgs, scales):
                cy = lognormal_pore_pdf(x, N, T, me, mp, mu, sigma, rg, t0)*scale
                cy_list.append(cy)
            ty = np.sum(cy_list, axis=0)
            return cy_list, ty

        def compute_mnp_curves(x, N, T, me, mp, poresize, t0):
            cy_list = []
            for rg, scale in zip(rgs, scales):
                rho = min(1, rg/poresize)
                ni_ = N * (1 - rho)**me
                ti_ = T * (1 - rho)**mp
                cy = robust_single_pore_pdf(x - t0, ni_, ti_)*scale
                cy_list.append(cy)
            ty = np.sum(cy_list, axis=0)
            return cy_list, ty

        cy_list, ty = compute_lnp_curves(x, N, T, me, mp, mu, sigma, t0)
        if use_ty_as_data:
            y = ty
        if y is not None:
            ax1.plot(x, y, label="data")
        for cy, rg in zip(cy_list, rgs):
            ax1.plot(x, cy, ":", label="rg=%.3g" % rg)
        if plot_ty:
            ax1.plot(x, ty, ":", label="model total", color="red")

        # ax1.legend()

        mode = compute_mode(mu, sigma)
        stdev = compute_stdev(mu, sigma)
        print("mode, stdev=", mode, stdev)
        stdev_percent = stdev/mode*100

        gen_params = np.array([N, T, t0, mode, stdev, stdev_percent])
        slider_specs = [    ("N", 0, 2000, gen_params[0]),
                            ("T", 0, 2, gen_params[1]),
                            ("t0",  -500, 500, gen_params[2]),
                            ("poresize", 10, 600, gen_params[3]),
                            ("stdev", 1, 100, gen_params[4]),
                            ("stdev_percent", 1, 50, gen_params[5]),
                            ]
        x_ = np.arange(-200, 600)
        cy_list, ty = compute_lnp_curves(x_, N, T, me, mp, mu, sigma, t0)

        t0_line = ax2.axvline(t0, color="red", label="t0")
        axt = ax2.twinx()
        axt.grid(False)
        excl_line, = axt.plot(trv, rgv, color="yellow")
        ax2.plot(x_, ty, color="gray", alpha=0.2)
        g_curves = []
        for cy, rg in zip(cy_list, rgs):
            curve, = ax2.plot(x_, cy, ":", label="rg=%.3g" % rg)
            g_curves.append(curve)
        curve, = ax2.plot(x_, ty, label="ln pore total", color="orange")
        g_curves.append(curve)
 
        if plot_mnp:
            poresize = compute_mode(mu, sigma)
            ty_mnp = compute_mnp_curves(x_, N, T, me, mp, poresize, t0)[1]
            mnp_curve, = ax2.plot(x_, ty_mnp, "-", label="mono pore total", color="green", alpha=0.5)

        ax2.legend()

        r = np.arange(0, 600)
        psd_y = distr_func(r, mu, sigma)
        ax3.plot(r, psd_y, color="gray", alpha=0.2)
        psd_curve, = ax3.plot(r, psd_y, alpha=0.5, label="Lognormal(%.3g, %.3g)" % (mu, sigma))
        psd_mode = ax3.axvline(mode, linestyle=":", color="green", label="mode")

        ymin, ymax = ax3.get_ylim()
        init_ymax = ymax*1.3
        ax3.set_ylim(ymin, init_ymax)
        ax3.legend(loc="upper left")    # need to update this if you show the legend

        psd_curve_label = get_labeltext_from_line(psd_curve)

        def slider_update(k, val):
            # print([k], "slider_update", val)
            # for eventson below see https://stackoverflow.com/questions/64420927/matplotlib-update-one-slider-based-on-change-in-another-slider
            gen_params[k] = val
            t0_line.set_xdata(gen_params[2])
            slider = None
            if k in [3, 4, 5]:
                if k in [3, 5]:
                    stdev = gen_params[5]/100*gen_params[3]
                    slider = sliders[4]
                    slider.eventson = False
                    slider.set_val(stdev)
                elif k == 4:
                    stdev = gen_params[4]
                    stdev_percent = gen_params[5]/gen_params[3]*100
                    slider = sliders[5]
                    slider.eventson = False
                    slider.set_val(stdev_percent)
            mu, sigma = compute_mu_sigma(gen_params[3], stdev)
            trv_ = compute_trvec(gen_params[0], gen_params[1], gen_params[2], me, mp, mu, sigma, rgv)
            excl_line.set_xdata(trv_)
            mu, sigma = compute_mu_sigma(gen_params[3], stdev)
            cy_list, ty = compute_lnp_curves(x_, gen_params[0], gen_params[1], me, mp, mu, sigma, gen_params[2])
            for curve, y_ in zip(g_curves, cy_list+[ty]):
                curve.set_ydata(y_)
            if plot_mnp:
                poresize = compute_mode(mu, sigma)
                ty_mnp = compute_mnp_curves(x_, gen_params[0], gen_params[1], me, mp, poresize, gen_params[2])[1]
                mnp_curve.set_ydata(ty_mnp)
                psd_mode.set_xdata(poresize)
            disty = distr_func(r, mu, sigma)
            psd_curve.set_ydata(disty)
            psd_curve_label.set_text("Lognormal(%.3g, %.3g)" % (mu, sigma))
            new_ymax = np.max(disty)*1.2
            ax3.set_ylim(ymin, max(init_ymax, new_ymax))
            fig.canvas.draw_idle()
            if slider is not None:
                slider.eventson = True

        slider_axes = []
        sliders = []
        for k, (label, valmin, valmax, valinit) in enumerate(slider_specs, start=0):
            ax_ = fig.add_axes([0.85, 0.7 - 0.08*k, 0.11, 0.03])
            slider  = Slider(ax_, label=label, valmin=valmin, valmax=valmax, valinit=valinit)
            slider.on_changed(lambda val, k_=k: slider_update(k_, val))
            slider_axes.append(ax_)
            sliders.append(slider)

        def show_curve_data(event):
            print("show_curve_data")
            for k, curve in enumerate(g_curves):
                x, y = curve.get_data()
                print([k], y[0:5])

        def reset(event):
            print("reset")
            for k, slider in enumerate(sliders):
                slider.reset()

        button_ax = plt.axes([0.85, 0.2, 0.12, 0.05])
        debug_btn = Button(button_ax, 'Reset', hovercolor='0.975')
        debug_btn.on_clicked(reset)

        fig.tight_layout()
        fig.subplots_adjust(right=0.8)
        ret = plt.show()
    
    return ret


"""
    XrayLrfConc.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right

def estimate_true_concentration(qv, D):
    i = bisect_right(qv, 0.02)

    cy_init = D[i,:]
    cy = cy_init
    cy2 = cy**2
    C_ = np.array([cy, cy2])

    C_list = [C_]

    for n in range(3):
        C_inv = np.linalg.pinv(C_)
        P_ = D @ C_inv
        P_inv = np.linalg.pinv(P_)
        C_ = P_inv @ D
        C_list.append(C_)

    return C_



# coding: utf-8
"""

    ElutionMapperAnimator.py

        recognition of peaks

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF

"""
import copy
import numpy                as np
import matplotlib.pyplot    as plt
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg
import matplotlib.animation as anm
from molass_legacy.KekLib.OurTkinter             import Tk, Dialog
from molass_legacy.KekLib.TkUtils                import is_low_resolution
from molass_legacy.KekLib.BasicUtils             import get_caller_module
from molass_legacy.KekLib.TkSupplements          import tk_set_icon_portable
from LightObjects           import AnyObject
from .MapperConstructor import create_mapper
from .ElutionMapperPlotter import ElutionMapperPlotter

class ElutionMapperAnimation( Dialog ):
    def __init__( self, parent, sd, mapper_hook=None ):
        self.parent         = parent
        self.serial_data    = sd

        mapper = create_mapper( parent, sd, mapper_hook=mapper_hook, anim_data=True )
        if mapper is None:
            pass

        self.mapper         = mapper
        self.three_d_guide  = False

        # self.caller_module  = get_caller_module( level=2 )

    def show( self ):
        Dialog.__init__( self, self.parent, "titile" )

    def body( self, body_frame ):   # overrides parent class method
        # tk_set_icon_portable( self, module=self.caller_module )

        cframe = Tk.Frame( body_frame )
        cframe.pack()

        plotter     = ElutionMapperPlotter( self, cframe, self.serial_data, anim_mode=True )
        animator    = ElutionMapperAnimator( plotter, self.mapper )
        animator.show()

    def update_plotter( self, i ):
        print( 'update_plotter', i )
        self.plotter.draw( clear= i != 0, animation_counter=i, mapper=self.mapper )

class ElutionMapperAnimator:
    def __init__( self, plotter, mapper ):
        self.plotter    = plotter
        self.mapper     = mapper
        self.num_frames = len( self.mapper.anim_data_list )
        print( 'num_frames=', self.num_frames )

    def show( self ):
        ani = anm.FuncAnimation( self.plotter.fig, self.update, fargs = (),
                                    interval = 100, frames = self.num_frames + 50 )

        self.plotter.mpl_canvas.draw()

    def update( self, i ):
        if i < self.num_frames:
            j = i
            done  = False
        else:
            j = -1
            done  = True
        mapper = self.mapper.anim_data_list[j]
        self.plotter.draw( clear= i != 0, animation_counter=i, mapper=mapper, animation_done=done )
